/**
 *  Copyright (c) 2008...2012 Ekkehard Gentz (Rosenheim, Germany) and others.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *  
 *  Contributors:
 *          Ekkehard Gentz     - initial contribution 
 *     
 *   more info: see folder [project]/copyright
 */
package org.appbus.bb.app.logging4bb;

import java.util.Enumeration;
import java.util.Hashtable;

import net.rim.device.api.command.Command;
import net.rim.device.api.command.CommandHandler;
import net.rim.device.api.command.CommandMetadata;
import net.rim.device.api.command.ReadOnlyCommandMetadata;
import net.rim.device.api.command.registrar.CategoryCollection;
import net.rim.device.api.command.registrar.CommandRequest;
import net.rim.device.api.command.registrar.RegistrationException;
import net.rim.device.api.command.registrar.RemoteCommandRegistrarConnection;
import net.rim.device.api.system.ApplicationManager;
import net.rim.device.api.system.Bitmap;
import net.rim.device.api.system.Characters;
import net.rim.device.api.ui.UiApplication;
import net.rim.device.api.ui.component.Dialog;
import net.rim.device.api.ui.image.Image;
import net.rim.device.api.ui.image.ImageFactory;
import net.rim.device.api.ui.menu.CommandItem;
import net.rim.device.api.util.LongHashtable;
import net.rim.device.api.util.StringProvider;

import org.appbus.bb.app.logging4bb.api.ILogger;
import org.appbus.bb.app.logging4bb.db.DatabaseOperationFailedException;
import org.appbus.bb.app.logging4bb.db.Logging2DatabaseWorker;
import org.appbus.bb.app.logging4bb.db.Logging4BBDatabaseManager;
import org.appbus.bb.app.logging4bb.db.TransmitLogEventsToServer;
import org.appbus.bb.app.logging4bb.logger.LoggerFactory;
import org.appbus.bb.app.logging4bb.model.LogEventTask;
import org.appbus.bb.app.logging4bb.socket.Logging2LogMonitorSocketWorker;

public class CommandFactory {
	
	private static final ILogger logger = LoggerFactory.getLogger(CommandFactory.class);
	
	// net.rim.device.api.command.registrar.RemoteCommandRegistrarConnection.CallbackContextCategoryKey
	// 0x174cf221a2db1c92L
	public static final Long CALLBACK_CONTEXT_CATEGORY_KEY = new Long(0x174cf221a2db1c92L);
	// net.rim.device.api.command.registrar.RemoteCommandRegistrarConnection.CallbackCommandCategoryKey
	// 0xdbf60f6bfc3d6474L
	public static final Long CALLBACK_COMMAND_CATEGORY_KEY = new Long(0xdbf60f6bfc3d6474L);
	// net.rim.device.api.command.registrar.RemoteCommandRegistrarConnection.CallbackContentKey
	// 0x8ee8a4348d8b84afL
	public static final Long CALLBACK_CONTENT_KEY = new Long(0x8ee8a4348d8b84afL);
	// net.rim.device.api.command.registrar.RemoteCommandRegistrarConnection.CallbackContextKey
	// 0xc7b172b7a680c283L
	public static final Long CALLBACK_CONTEXT_KEY = new Long(0xc7b172b7a680c283L);
	
	
	// workaround if registration fails
	// net.rim.device.api.command.registrar.RemoteCommandRegistrarConnection.UseGlobalEventAsCallBackWorkaroundKey;
	// 0x888cf272dd57a3b8L
	public static final Long USE_GLOBAL_EVENT_AS_WORKAOUND_KEY = new Long(0x888cf272dd57a3b8L);
	
	private static String COMMAND_ID_DELIMITER = "@";
	private static String LEFT_ALIGNMENT = "left";
	
	/**
	 * at startup of application remote commands should be registered
	 */
	public static void registerRemoteCommands() {
		logger.debug("registering remote commands");
		RemoteCommandRegistrarConnection connection = new RemoteCommandRegistrarConnection();
		// if MetaData is null we initialize the 'normal' Command

		//
		if (logEventFromRemoteClientCommandMetaData == null) {
			logEventFromRemoteClientCommand();
		}
		try {
			connection.registerCommand(
					logEventFromRemoteClientHandler, 
					logEventFromRemoteClientCommandMetaData);
		} catch (RegistrationException e) {
			// add timestamp to ID
			logger.debug("workaround", "added timestamp to Command ID: "+logEventFromRemoteClientCommandID);
			// TODO bug in remote command registry, fixed in 7.x
			logEventFromRemoteClientCommandID += COMMAND_ID_DELIMITER;
			logEventFromRemoteClientCommandID += System.currentTimeMillis();
			// reset Command --> creates new Metadata
			logEventFromRemoteClientCmd = null;
			logEventFromRemoteClientCommand();
			try {
				connection.registerCommand(
						logEventFromRemoteClientCommandHandler(), 
						logEventFromRemoteClientCommandMetaData);
			} catch (RegistrationException e2) {
				// now something goes really wrong
			}
		} catch (Throwable e) {
			// TODO: log
		}
		
		//
		if (refreshPreferencesFromDatabaseCommandMetaData == null) {
			refreshPreferencesFromDatabaseCommand();
		}
		try {
			connection.registerCommand(
					refreshPreferencesFromDatabaseHandler, 
					refreshPreferencesFromDatabaseCommandMetaData);
		} catch (RegistrationException e) {
			// add timestamp to ID
			logger.debug("workaround", "added timestamp to Command ID: "+refreshPreferencesFromDatabaseCommandID);
			// TODO bug in remote command registry, fixed in 7.x
			refreshPreferencesFromDatabaseCommandID += COMMAND_ID_DELIMITER;
			refreshPreferencesFromDatabaseCommandID += System.currentTimeMillis();
			// reset Command --> creates new Metadata
			refreshPreferencesFromDatabaseCmd = null;
			refreshPreferencesFromDatabaseCommand();
			try {
				connection.registerCommand(
						refreshPreferencesFromDatabaseCommandHandler(), 
						refreshPreferencesFromDatabaseCommandMetaData);
			} catch (RegistrationException e2) {
				// now something goes really wrong
			}
		} catch (Throwable e) {
			// TODO: log
		}
		
		//
		if (pauseLoggerCommandMetaData == null) {
			pauseLoggerCommand();
		}
		try {
			connection.registerCommand(
					pauseLoggerHandler, 
					pauseLoggerCommandMetaData);
		} catch (RegistrationException e) {
			// add timestamp to ID
			logger.debug("workaround", "added timestamp to Command ID: "+pauseLoggerCommandID);
			// TODO bug in remote command registry, fixed in 7.x
			pauseLoggerCommandID += COMMAND_ID_DELIMITER;
			pauseLoggerCommandID += System.currentTimeMillis();
			// reset Command --> creates new Metadata
			pauseLoggerCmd = null;
			pauseLoggerCommand();
			try {
				connection.registerCommand(
						pauseLoggerCommandHandler(), 
						pauseLoggerCommandMetaData);
			} catch (RegistrationException e2) {
				// now something goes really wrong
			}
		} catch (Throwable e) {
			// TODO: log
		}
		
		//
		if (resumeLoggerCommandMetaData == null) {
			resumeLoggerCommand();
		}
		try {
			connection.registerCommand(
					resumeLoggerHandler, 
					resumeLoggerCommandMetaData);
		} catch (RegistrationException e) {
			// add timestamp to ID
			logger.debug("workaround", "added timestamp to Command ID: "+resumeLoggerCommandID);
			resumeLoggerCommandID += COMMAND_ID_DELIMITER;
			resumeLoggerCommandID += System.currentTimeMillis();
			// reset Command --> creates new Metadata
			resumeLoggerCmd = null;
			resumeLoggerCommand();
			try {
				connection.registerCommand(
						resumeLoggerCommandHandler(), 
						resumeLoggerCommandMetaData);
			} catch (RegistrationException e2) {
				// now something goes really wrong
			}
		} catch (Throwable e) {
			// TODO: log
		}
		
		//
		//
		if (startTransmissionToServerCommandMetaData == null) {
			startTransmissionToServerCommand();
		}
		try {
			connection.registerCommand(
					startTransmissionToServerHandler, 
					startTransmissionToServerCommandMetaData);
		} catch (RegistrationException e) {
			// add timestamp to ID
			logger.debug("workaround", "added timestamp to Command ID: "+startTransmissionToServerCommandID);
			startTransmissionToServerCommandID += COMMAND_ID_DELIMITER;
			startTransmissionToServerCommandID += System.currentTimeMillis();
			// reset Command --> creates new Metadata
			startTransmissionToServerCmd = null;
			startTransmissionToServerCommand();
			try {
				connection.registerCommand(
						startTransmissionToServerCommandHandler(), 
						startTransmissionToServerCommandMetaData);
			} catch (RegistrationException e2) {
				// now something goes really wrong
			}
		} catch (Throwable e) {
			// TODO: log
		}
		
		//
		if (analyzeLoggedDataCommandMetaData == null) {
			analyzeLoggedDataCommand();
		}
		try {
			connection.registerCommand(
					analyzeLoggedDataHandler, 
					analyzeLoggedDataCommandMetaData);
		} catch (RegistrationException e) {
			// add timestamp to ID
			logger.debug("workaround", "added timestamp to Command ID: "+analyzeLoggedDataCommandID);
			analyzeLoggedDataCommandID += COMMAND_ID_DELIMITER;
			analyzeLoggedDataCommandID += System.currentTimeMillis();
			// reset Command --> creates new Metadata
			analyzeLoggedDataCmd = null;
			analyzeLoggedDataCommand();
			try {
				connection.registerCommand(
						analyzeLoggedDataCommandHandler(), 
						analyzeLoggedDataCommandMetaData);
			} catch (RegistrationException e2) {
				// now something goes really wrong
			}
		} catch (Throwable e) {
			// TODO: log
		}
		
		//
		if (deleteTransmittedLogEventsCommandMetaData == null) {
			deleteTransmittedLogEventsCommand();
		}
		try {
			connection.registerCommand(
					deleteTransmittedLogEventsHandler, 
					deleteTransmittedLogEventsCommandMetaData);
		} catch (RegistrationException e) {
			// add timestamp to ID
			logger.debug("workaround", "added timestamp to Command ID "+deleteTransmittedLogEventsCommandID);
			deleteTransmittedLogEventsCommandID += COMMAND_ID_DELIMITER;
			deleteTransmittedLogEventsCommandID += System.currentTimeMillis();
			// reset Command --> creates new Metadata
			deleteTransmittedLogEventsCmd = null;
			deleteTransmittedLogEventsCommand();
			try {
				connection.registerCommand(
						deleteTransmittedLogEventsCommandHandler(), 
						deleteTransmittedLogEventsCommandMetaData);
			} catch (RegistrationException e2) {
				// now something goes really wrong
			}
		} catch (Throwable e) {
			// TODO: log
		}
		
	}
	
	// remote commands to be executed
	
	private static void notifyUI(final String message){
		UiApplication.getUiApplication().invokeLater(
		new Runnable()		{
			public void run() {
				Dialog.inform(message);
			};
		}
		
		);
	}
	
	
	// callback commands
	/**
	 * executes the remote command, from where we were called
	 * callbackCommandName and Categories are needed to identify the remote command
	 * 
	 * if the origin command had a context, we'll give it back -
	 * could be a KEY or a screen object etc.
	 * 
	 * if the result of the command is anything, it goes back as content,
	 * could be an ErrorMessage or a result as [] or Vector etc.
	 * 
	 * context and content can be null
	 * 
	 * if context is a HashTable itself, all key/value will be added
	 * 
	 */
	public static void logging4bbCallbackCommandRemoteExecute(
				String callbackCommandName,
				String[] logging4bbCallbackCommandCategories, 
				String[] logging4bbCallbackContextCategories,
				Object context,
				Object content){
		CommandContextHashtable hashtable = new CommandContextHashtable(callbackCommandName);
		if (context != null) {
			if (context instanceof Hashtable) {
				// we copy the keys and values
				Hashtable contextHashtable = (Hashtable) context;
				Enumeration en = contextHashtable.keys();
				while (en.hasMoreElements()) {
					Object key = en.nextElement();
					Object value = contextHashtable.get(key);
					if (value != null) {
						hashtable.put(key, value);
					}
				}
			} else {
				hashtable.put(CALLBACK_CONTEXT_KEY, context);
			}
		}
		if (content != null) {
			hashtable.put(CALLBACK_CONTENT_KEY, content);
		}
		CommandRequest request = new CommandRequest();
    	request.setCommandCategories(new CategoryCollection(logging4bbCallbackCommandCategories));
    	request.setContextCategories(new CategoryCollection(logging4bbCallbackContextCategories));
    	request.setAndContextCategories(true);
    	RemoteCommandRegistrarConnection connection = new RemoteCommandRegistrarConnection();
    	Command[] logging4bbCallbackCmds = connection.getCommands(request);
    	Command logging4bbCallbackCmd = getNewestCommand(logging4bbCallbackCmds);
    	if (logging4bbCallbackCmd == null) {
    		notifyUI("No Remote Command match Categories - is Logging4BBManager running ?");
			return;
		}
    	logging4bbCallbackCmd.execute(hashtable);
	}
	
	// helper workaround bugfix 6.0
	// works on 7.0 because you can unregister commands
	
	private static Command getNewestCommand(Command[] commands){
		if (commands == null || commands.length == 0) {
			return null;
		}
		if (commands.length == 1) {
			return commands[0];
		}
		long highScoreTimestamp = Long.MIN_VALUE;
		int indexOfHighScoreCommand = 0;
		for (int i = 0; i < commands.length; i++) {
			String commandId = commands[i].getId();
			int posOfDelimiter = commandId.indexOf(COMMAND_ID_DELIMITER);
			if (posOfDelimiter > 0) {
				try {
					String timestampAsString = commandId.substring(posOfDelimiter+1);
					long timestamp = Long.parseLong(timestampAsString);
					if (timestamp > highScoreTimestamp) {
						highScoreTimestamp = timestamp;
						indexOfHighScoreCommand = i;
					}
				} catch (IndexOutOfBoundsException e) {
					// ignore
				} catch (NumberFormatException e) {
					// ignore
				}
			} 
		}
		logger.debug("OS 6 Workaround", "found commands: "+commands.length);
		return commands[indexOfHighScoreCommand];
	}
	
	// my own commands
	
	// refreshPreferencesFromDatabase
	// step1: stops (pause) all enabled logger
	// step 2: updates transmitted values into database Preferences table
	// step 3: restarts all enabled logger
	private static Command refreshPreferencesFromDatabaseCmd;
	private static CommandItem refreshPreferencesFromDatabaseCmdItem;
	private static String refreshPreferencesFromDatabaseCommandID = "refreshPreferencesFromDatabase";
	private static CommandMetadata refreshPreferencesFromDatabaseCommandMetaData;
	private static String[] refreshPreferencesFromDatabaseCommandCategories = { "refreshPreferencesFromDatabase" };
	private static String[] refreshPreferencesFromDatabaseContextCategories = { Logging4BBApp.MY_APPLICATION_CONTEXT_CATEGORY };
	private static String refreshPreferencesFromDatabaseCommandBitmapIcon = "refreshPreferencesFromDatabase_32.png";
	private static Image refreshPreferencesFromDatabaseImage;
	private static Bitmap refreshPreferencesFromDatabaseBitmap;
	private static CommandHandler refreshPreferencesFromDatabaseHandler;

	private static CommandHandler refreshPreferencesFromDatabaseCommandHandler() {
		if (refreshPreferencesFromDatabaseHandler == null) {
			refreshPreferencesFromDatabaseHandler = new CommandHandler() {
				public void execute(ReadOnlyCommandMetadata metadata,
						Object context) {
					if (canExecute(metadata, context)) {
						if (context instanceof Hashtable) {
							logger.debug("refresh Preferences","pause existing logger");
							pauseLoggerCommand().execute(new Hashtable());
							Logging4BBDatabaseManager.refreshPreferences((Hashtable)context);
						}
					}
					return;
				}

				public boolean canExecute(ReadOnlyCommandMetadata metadata,
						Object context) {
//					if (context instanceof IExecutionContext) {
//						return ((IExecutionContext) context).canExecute();
//					}
					// TODO insert code to test if Command can be executed
					return true;
				}
			};
		}
		return refreshPreferencesFromDatabaseHandler;
	}

	public static Command refreshPreferencesFromDatabaseCommand() {
		if (refreshPreferencesFromDatabaseCmd == null) {
			refreshPreferencesFromDatabaseCommandMetaData = new CommandMetadata(refreshPreferencesFromDatabaseCommandID);
			refreshPreferencesFromDatabaseCommandMetaData.set(CommandMetadata.BITMAP_ICON,
					refreshPreferencesFromDatabaseCommandBitmapIcon);
			refreshPreferencesFromDatabaseCommandMetaData
					.set(CommandMetadata.ALIGNMENT, LEFT_ALIGNMENT);
			refreshPreferencesFromDatabaseCommandMetaData.setCommandCategories(new CategoryCollection(
					refreshPreferencesFromDatabaseCommandCategories));
			refreshPreferencesFromDatabaseCommandMetaData.setContextCategories(new CategoryCollection(
					refreshPreferencesFromDatabaseContextCategories));
			refreshPreferencesFromDatabaseCmd = new Command(refreshPreferencesFromDatabaseCommandHandler(),
					refreshPreferencesFromDatabaseCommandMetaData);
		}
		return refreshPreferencesFromDatabaseCmd;
	}

	public static CommandItem refreshPreferencesFromDatabaseCommandItem() {
		if (refreshPreferencesFromDatabaseCmdItem == null) {
			refreshPreferencesFromDatabaseCmdItem = new CommandItem(refreshPreferencesFromDatabaseCommand());
			refreshPreferencesFromDatabaseCmdItem.setIcon(refreshPreferencesFromDatabaseCommandImage());
			// refreshPreferencesFromDatabaseCmdItem.setText(ActionStringProviderFactory.refreshPreferencesFromDatabase_Action());
			refreshPreferencesFromDatabaseCmdItem.setText(new StringProvider("refreshPreferencesFromDatabase"));
		}
		return refreshPreferencesFromDatabaseCmdItem;
	}

	public static Bitmap refreshPreferencesFromDatabaseCommandBitmap() {
		if (refreshPreferencesFromDatabaseBitmap == null) {
			refreshPreferencesFromDatabaseBitmap = Bitmap.getBitmapResource(refreshPreferencesFromDatabaseCommandBitmapIcon);
		}
		return refreshPreferencesFromDatabaseBitmap;
	}

	public static Image refreshPreferencesFromDatabaseCommandImage() {
		if (refreshPreferencesFromDatabaseImage == null) {
			refreshPreferencesFromDatabaseImage = ImageFactory.createImage(refreshPreferencesFromDatabaseCommandBitmap());
		}
		return refreshPreferencesFromDatabaseImage;
	}
	
	// pauseTracker
	private static Command pauseLoggerCmd;
	private static CommandItem pauseLoggerCmdItem;
	private static String pauseLoggerCommandID = "pauseLogger";
	private static CommandMetadata pauseLoggerCommandMetaData;
	private static String[] pauseLoggerCommandCategories = { "pauseLogger" };
	private static String[] pauseLoggerContextCategories = { Logging4BBApp.MY_APPLICATION_CONTEXT_CATEGORY };
	private static String pauseLoggerCommandBitmapIcon = "pauseLogger_32.png";
	private static Image pauseLoggerImage;
	private static Bitmap pauseLoggerBitmap;
	private static CommandHandler pauseLoggerHandler;

	private static CommandHandler pauseLoggerCommandHandler() {
		if (pauseLoggerHandler == null) {
			pauseLoggerHandler = new CommandHandler() {
				public void execute(ReadOnlyCommandMetadata metadata,
						Object context) {
					if (canExecute(metadata, context)) {
						// simple command to restart the loggers
						// as explained in HelloLogger example
						if (Logging4BBApp.logToDB()) {
							Logging2DatabaseWorker.pauseWorker();
						}
						if (Logging4BBApp.logToLogmonitor()) {
							Logging2LogMonitorSocketWorker.pauseWorker();
						}
						if (Logging4BBApp.logToWebsocket()) {
							// TODO websockets
						}
						// callbacks etc are only used by LogMonitor
						// available at AppWorld
						boolean isCallbackContext = false;
						Hashtable contextHashtable = null;
						StringProvider commandCategory = null;
						StringProvider contextCategory = null;
						if (context instanceof Hashtable) {
							contextHashtable = (Hashtable) context;
							if (contextHashtable.size() >= 2) {
								Object o = contextHashtable.get(CALLBACK_COMMAND_CATEGORY_KEY);
								if (o instanceof StringProvider) {
									commandCategory = (StringProvider)o;
								}
								o = contextHashtable.get(CALLBACK_CONTEXT_CATEGORY_KEY);
								if (o instanceof StringProvider) {
									contextCategory = (StringProvider)o;
								}
								isCallbackContext = (commandCategory != null && contextCategory != null);
							}
							if (isCallbackContext) {
								StringProvider content = new StringProvider("Logger PAUSE");
								logging4bbCallbackCommandRemoteExecute(
										"logger4bbCallback",
										new String[]{commandCategory.toString()},
										new String[] {contextCategory.toString()},
										null, // no context to be sent back
										content // resulting message
								);
							}
						}
					}
					return;
				}

				public boolean canExecute(ReadOnlyCommandMetadata metadata,
						Object context) {

					// TODO insert code to test if Command can be executed
					return true;
				}
			};
		}
		return pauseLoggerHandler;
	}

	public static Command pauseLoggerCommand() {
		if (pauseLoggerCmd == null) {
			pauseLoggerCommandMetaData = new CommandMetadata(pauseLoggerCommandID);
			pauseLoggerCommandMetaData.set(CommandMetadata.BITMAP_ICON,
					pauseLoggerCommandBitmapIcon);
			pauseLoggerCommandMetaData
					.set(CommandMetadata.ALIGNMENT, LEFT_ALIGNMENT);
			pauseLoggerCommandMetaData.setCommandCategories(new CategoryCollection(
					pauseLoggerCommandCategories));
			pauseLoggerCommandMetaData.setContextCategories(new CategoryCollection(
					pauseLoggerContextCategories));
			pauseLoggerCmd = new Command(pauseLoggerCommandHandler(),
					pauseLoggerCommandMetaData);
		}
		return pauseLoggerCmd;
	}

	public static CommandItem pauseLoggerCommandItem() {
		if (pauseLoggerCmdItem == null) {
			pauseLoggerCmdItem = new CommandItem(pauseLoggerCommand());
			pauseLoggerCmdItem.setIcon(pauseLoggerCommandImage());
			//pauseLoggerCmdItem.setText(ActionStringProviderFactory.pauseTracker_Action());
			pauseLoggerCmdItem.setText(new StringProvider("pauseLogger"));
		}
		return pauseLoggerCmdItem;
	}

	public static Bitmap pauseLoggerCommandBitmap() {
		if (pauseLoggerBitmap == null) {
			pauseLoggerBitmap = Bitmap.getBitmapResource(pauseLoggerCommandBitmapIcon);
		}
		return pauseLoggerBitmap;
	}

	public static Image pauseLoggerCommandImage() {
		if (pauseLoggerImage == null) {
			pauseLoggerImage = ImageFactory.createImage(pauseLoggerCommandBitmap());
		}
		return pauseLoggerImage;
	}
	
	// resumeLogger
	private static Command resumeLoggerCmd;
	private static CommandItem resumeLoggerCmdItem;
	private static String resumeLoggerCommandID = "resumeLogger";
	private static CommandMetadata resumeLoggerCommandMetaData;
	private static String[] resumeLoggerCommandCategories = { "resumeLogger" };
	private static String[] resumeLoggerContextCategories = { Logging4BBApp.MY_APPLICATION_CONTEXT_CATEGORY };
	private static String resumeLoggerCommandBitmapIcon = "resumeLogger_32.png";
	private static Image resumeLoggerImage;
	private static Bitmap resumeLoggerBitmap;
	private static CommandHandler resumeLoggerHandler;

	private static CommandHandler resumeLoggerCommandHandler() {
		if (resumeLoggerHandler == null) {
			resumeLoggerHandler = new CommandHandler() {
				public void execute(ReadOnlyCommandMetadata metadata,
						Object context) {
					if (canExecute(metadata, context)) {
						// simple command to restart the loggers
						// as explained in HelloLogger example
						if (Logging4BBApp.logToDB()) {
							Logging2DatabaseWorker.restartWorker();
						}
						if (Logging4BBApp.logToLogmonitor()) {
							Logging2LogMonitorSocketWorker.restartWorker();
						}
						if (Logging4BBApp.logToWebsocket()) {
							// TODO websockets
						}
						// callbacks etc are only used by LogMonitor
						// available at AppWorld
						boolean isCallbackContext = false;
						Hashtable contextHashtable = null;
						StringProvider commandCategory = null;
						StringProvider contextCategory = null;
						if (context instanceof Hashtable) {
							contextHashtable = (Hashtable) context;
							if (contextHashtable.size() >= 2) {
								Object o = contextHashtable.get(CALLBACK_COMMAND_CATEGORY_KEY);
								if (o instanceof StringProvider) {
									commandCategory = (StringProvider)o;
								}
								o = contextHashtable.get(CALLBACK_CONTEXT_CATEGORY_KEY);
								if (o instanceof StringProvider) {
									contextCategory = (StringProvider)o;
								}
								isCallbackContext = (commandCategory != null && contextCategory != null);
							}
							if (isCallbackContext) {
								StringProvider content = new StringProvider("Logger RESUME");
								logging4bbCallbackCommandRemoteExecute(
										"logging4bbCallback",
										new String[]{commandCategory.toString()},
										new String[] {contextCategory.toString()},
										null, // no context to send back
										content // message to send back
								);
							}
						}
					}
					return;
				}

				public boolean canExecute(ReadOnlyCommandMetadata metadata,
						Object context) {

					// TODO insert code to test if Command can be executed
					return true;
				}
			};
		}
		return resumeLoggerHandler;
	}

	public static Command resumeLoggerCommand() {
		if (resumeLoggerCmd == null) {
			resumeLoggerCommandMetaData = new CommandMetadata(resumeLoggerCommandID);
			resumeLoggerCommandMetaData.set(CommandMetadata.BITMAP_ICON,
					resumeLoggerCommandBitmapIcon);
			resumeLoggerCommandMetaData
					.set(CommandMetadata.ALIGNMENT, LEFT_ALIGNMENT);
			resumeLoggerCommandMetaData.setCommandCategories(new CategoryCollection(
					resumeLoggerCommandCategories));
			resumeLoggerCommandMetaData.setContextCategories(new CategoryCollection(
					resumeLoggerContextCategories));
			resumeLoggerCmd = new Command(resumeLoggerCommandHandler(),
					resumeLoggerCommandMetaData);
		}
		return resumeLoggerCmd;
	}

	public static CommandItem resumeTrackerCommandItem() {
		if (resumeLoggerCmdItem == null) {
			resumeLoggerCmdItem = new CommandItem(resumeLoggerCommand());
			resumeLoggerCmdItem.setIcon(resumeLoggerCommandImage());
			//resumeLoggerCmdItem.setText(ActionStringProviderFactory.resumeTracker_Action());
			resumeLoggerCmdItem.setText(new StringProvider("resumeLogger"));
		}
		return resumeLoggerCmdItem;
	}

	public static Bitmap resumeLoggerCommandBitmap() {
		if (resumeLoggerBitmap == null) {
			resumeLoggerBitmap = Bitmap.getBitmapResource(resumeLoggerCommandBitmapIcon);
		}
		return resumeLoggerBitmap;
	}

	public static Image resumeLoggerCommandImage() {
		if (resumeLoggerImage == null) {
			resumeLoggerImage = ImageFactory.createImage(resumeLoggerCommandBitmap());
		}
		return resumeLoggerImage;
	}
	
	
	// startTransmissionToServer
	private static Command startTransmissionToServerCmd;
	private static CommandItem startTransmissionToServerCmdItem;
	private static String startTransmissionToServerCommandID = "startTransmissionToServer";
	private static CommandMetadata startTransmissionToServerCommandMetaData;
	private static String[] startTransmissionToServerCommandCategories = { "startTransmissionToServer" };
	private static String[] startTransmissionToServerContextCategories = { Logging4BBApp.MY_APPLICATION_CONTEXT_CATEGORY };
	private static String startTransmissionToServerCommandBitmapIcon = "startTransmissionToServer_32.png";
	private static Image startTransmissionToServerImage;
	private static Bitmap startTransmissionToServerBitmap;
	private static CommandHandler startTransmissionToServerHandler;

	private static CommandHandler startTransmissionToServerCommandHandler() {
		if (startTransmissionToServerHandler == null) {
			startTransmissionToServerHandler = new CommandHandler() {
				public void execute(ReadOnlyCommandMetadata metadata,
						Object context) {
					if (canExecute(metadata, context)) {
						new Runnable() {
							// Hint: we must use a runnable to break the chain:
							// this command creates a thread, which itself starts this command
							// at the end again if all was successfully
							// to free the references the call-chain must be decoupled !
							public void run() {
								// Query for not transmitted
								LongHashtable results;
								try {
									results = Logging4BBDatabaseManager.selectRecordToTransmit();
								} catch (DatabaseOperationFailedException e) {
									// TODO LOG
									return;
								}
								// loop thru results
								if (results == null || results.size() == 0) {
									// TODO LOG
									// all transmitted - now the app would wait until
									// unconnect / connect to WLAN again
									// to avoid a holw if for long time inside WLAN:
									// set intervall to be restarted later again if still in WLAN
									Logging4BBApp.getInstance().setLastTransmitInsideWlan(System.currentTimeMillis());
									return;
								}
								// ASYNC THREAD
								// double check: test if already active
								// can't start if running - otherwise we'll get unwanted duplicates
								if (Logging4BBApp.getInstance().isTransmitLogEventsToServerThreadRunning()) {
									// TODO log or UI feedback
									return;
								}
								// let it GO
								Logging4BBApp.getInstance().setTransmitLogEventsToServerThreadRunning(true);
								new TransmitLogEventsToServer(results).start();
							}
						}.run();
					}
					return;
				}

				public boolean canExecute(ReadOnlyCommandMetadata metadata,
						Object context) {
					if (Logging4BBApp.getInstance().isTransmitLogEventsToServerThreadRunning()) {
						return false;
					}
					return true;
				}
			};
		}
		return startTransmissionToServerHandler;
	}

	public static Command startTransmissionToServerCommand() {
		if (startTransmissionToServerCmd == null) {
			startTransmissionToServerCommandMetaData = new CommandMetadata(startTransmissionToServerCommandID);
			startTransmissionToServerCommandMetaData.set(CommandMetadata.BITMAP_ICON,
					startTransmissionToServerCommandBitmapIcon);
			startTransmissionToServerCommandMetaData
					.set(CommandMetadata.ALIGNMENT, LEFT_ALIGNMENT);
			startTransmissionToServerCommandMetaData.setCommandCategories(new CategoryCollection(
					startTransmissionToServerCommandCategories));
			startTransmissionToServerCommandMetaData.setContextCategories(new CategoryCollection(
					startTransmissionToServerContextCategories));
			startTransmissionToServerCmd = new Command(startTransmissionToServerCommandHandler(),
					startTransmissionToServerCommandMetaData);
		}
		return startTransmissionToServerCmd;
	}

	public static CommandItem startTransmissionToServerCommandItem() {
		if (startTransmissionToServerCmdItem == null) {
			startTransmissionToServerCmdItem = new CommandItem(startTransmissionToServerCommand());
			startTransmissionToServerCmdItem.setIcon(startTransmissionToServerCommandImage());
			// startTransmissionToServerCmdItem.setText(ActionStringProviderFactory.startTransmissionToServer_Action());
			startTransmissionToServerCmdItem.setText(new StringProvider("startTransmissionToServer"));
		}
		return startTransmissionToServerCmdItem;
	}

	public static Bitmap startTransmissionToServerCommandBitmap() {
		if (startTransmissionToServerBitmap == null) {
			startTransmissionToServerBitmap = Bitmap.getBitmapResource(startTransmissionToServerCommandBitmapIcon);
		}
		return startTransmissionToServerBitmap;
	}

	public static Image startTransmissionToServerCommandImage() {
		if (startTransmissionToServerImage == null) {
			startTransmissionToServerImage = ImageFactory.createImage(startTransmissionToServerCommandBitmap());
		}
		return startTransmissionToServerImage;
	}
	
	// analyzeLoggedData
	private static Command analyzeLoggedDataCmd;
	private static CommandItem analyzeLoggedDataCmdItem;
	private static String analyzeLoggedDataCommandID = "analyzeLoggedData";
	private static CommandMetadata analyzeLoggedDataCommandMetaData;
	private static String[] analyzeLoggedDataCommandCategories = { "analyzeLoggedData" };
	private static String[] analyzeLoggedDataContextCategories = { Logging4BBApp.MY_APPLICATION_CONTEXT_CATEGORY };
	private static String analyzeLoggedDataCommandBitmapIcon = "analyzeLoggedData_32.png";
	private static Image analyzeLoggedDataImage;
	private static Bitmap analyzeLoggedDataBitmap;
	private static CommandHandler analyzeLoggedDataHandler;

	private static CommandHandler analyzeLoggedDataCommandHandler() {
		if (analyzeLoggedDataHandler == null) {
			analyzeLoggedDataHandler = new CommandHandler() {
				public void execute(ReadOnlyCommandMetadata metadata,
						Object context) {
					if (canExecute(metadata, context)) {
						boolean isCallbackContext = false;
						long useGlobalEventAsCallBackWorkaroundGuid = Long.MIN_VALUE;
						Hashtable contextHashtable = null;
						StringProvider commandCategory = null;
						StringProvider contextCategory = null;
						Object callbackContext = null;
						if (context instanceof Hashtable) {
							contextHashtable = (Hashtable) context;
							if (contextHashtable.size() >= 2) {
								Object o = contextHashtable.get(CALLBACK_COMMAND_CATEGORY_KEY);
								if (o instanceof StringProvider) {
									commandCategory = (StringProvider)o;
								}
								o = contextHashtable.get(CALLBACK_CONTEXT_CATEGORY_KEY);
								if (o instanceof StringProvider) {
									contextCategory = (StringProvider)o;
								}
								isCallbackContext = (commandCategory != null && contextCategory != null);
								callbackContext = contextHashtable.get(CALLBACK_CONTEXT_KEY); // can be null
								if (callbackContext != null) {
									o = contextHashtable.get(USE_GLOBAL_EVENT_AS_WORKAOUND_KEY);
									if (o instanceof Long) {
										useGlobalEventAsCallBackWorkaroundGuid = ((Long)o).longValue();
									}
								}
							}
						}
						int countRecords = 0;
						try {
							countRecords = Logging4BBDatabaseManager.countAllLogEvents();
						} catch (DatabaseOperationFailedException e) {
							// TODO log
						}
						int countNotTransmitted = 0;
						try {
							countNotTransmitted = Logging4BBDatabaseManager.countNotTransmittedLogEvents();
						} catch (DatabaseOperationFailedException e) {
							// TODO log
						}
						StringBuffer sb = new StringBuffer();
						sb.append(Logging4BBDatabaseManager.TABLE_LOG_EVENT);
						sb.append(Characters.COLON);
						sb.append(Characters.SPACE);
						sb.append(countRecords);
						sb.append(Characters.SPACE);
						sb.append(Characters.LEFT_PARENTHESIS);
						sb.append("not transmitted: ");
						sb.append(countNotTransmitted);
						sb.append(Characters.RIGHT_PARENTHESIS);
						sb.append("\n");
						if (isCallbackContext) {
							if (useGlobalEventAsCallBackWorkaroundGuid != Long.MIN_VALUE) {
								// workaround if registration failed
								Hashtable hashtable = new Hashtable(2);
								hashtable.put(CALLBACK_CONTENT_KEY, sb.toString());
								hashtable.put(CALLBACK_CONTEXT_KEY, callbackContext);
								ApplicationManager.getApplicationManager()
									.postGlobalEvent(
											useGlobalEventAsCallBackWorkaroundGuid, 
											-1, -1,
											hashtable, 
											null);
							} else {
								logging4bbCallbackCommandRemoteExecute(
										"logging4bbCallback",
										new String[]{commandCategory.toString()},
										new String[] {contextCategory.toString()},
										callbackContext, // context to be send back
										sb.toString() // error message or vector with results
								);
							}
						}
					}
					return;
				}

				public boolean canExecute(ReadOnlyCommandMetadata metadata,
						Object context) {

					// TODO insert code to test if Command can be executed
					return true;
				}
			};
		}
		return analyzeLoggedDataHandler;
	}

	public static Command analyzeLoggedDataCommand() {
		if (analyzeLoggedDataCmd == null) {
			analyzeLoggedDataCommandMetaData = new CommandMetadata(analyzeLoggedDataCommandID);
			analyzeLoggedDataCommandMetaData.set(CommandMetadata.BITMAP_ICON,
					analyzeLoggedDataCommandBitmapIcon);
			analyzeLoggedDataCommandMetaData
					.set(CommandMetadata.ALIGNMENT, LEFT_ALIGNMENT);
			analyzeLoggedDataCommandMetaData.setCommandCategories(new CategoryCollection(
					analyzeLoggedDataCommandCategories));
			analyzeLoggedDataCommandMetaData.setContextCategories(new CategoryCollection(
					analyzeLoggedDataContextCategories));
			analyzeLoggedDataCmd = new Command(analyzeLoggedDataCommandHandler(),
					analyzeLoggedDataCommandMetaData);
		}
		return analyzeLoggedDataCmd;
	}

	public static CommandItem analyzeLoggedDataCommandItem() {
		if (analyzeLoggedDataCmdItem == null) {
			analyzeLoggedDataCmdItem = new CommandItem(analyzeLoggedDataCommand());
			analyzeLoggedDataCmdItem.setIcon(analyzeLoggedDataCommandImage());
			//analyzeLoggedDataCmdItem.setText(ActionStringProviderFactory.analyzeTrackedData_Action());
			analyzeLoggedDataCmdItem.setText(new StringProvider("analyzeLoggedData"));
		}
		return analyzeLoggedDataCmdItem;
	}

	public static Bitmap analyzeLoggedDataCommandBitmap() {
		if (analyzeLoggedDataBitmap == null) {
			analyzeLoggedDataBitmap = Bitmap.getBitmapResource(analyzeLoggedDataCommandBitmapIcon);
		}
		return analyzeLoggedDataBitmap;
	}

	public static Image analyzeLoggedDataCommandImage() {
		if (analyzeLoggedDataImage == null) {
			analyzeLoggedDataImage = ImageFactory.createImage(analyzeLoggedDataCommandBitmap());
		}
		return analyzeLoggedDataImage;
	}
		
	// deleteTransmittedLogEvents
	private static Command deleteTransmittedLogEventsCmd;
	private static CommandItem deleteTransmittedLogEventsCmdItem;
	private static String deleteTransmittedLogEventsCommandID = "deleteTransmittedLogEvents";
	private static CommandMetadata deleteTransmittedLogEventsCommandMetaData;
	private static String[] deleteTransmittedLogEventsCommandCategories = { "deleteTransmittedLogEvents" };
	private static String[] deleteTransmittedLogEventsContextCategories = { Logging4BBApp.MY_APPLICATION_CONTEXT_CATEGORY };
	private static String deleteTransmittedLogEventsCommandBitmapIcon = "deleteTransmittedLogEvents_32.png";
	private static Image deleteTransmittedLogEventsImage;
	private static Bitmap deleteTransmittedLogEventsBitmap;
	private static CommandHandler deleteTransmittedLogEventsHandler;

	private static CommandHandler deleteTransmittedLogEventsCommandHandler() {
		if (deleteTransmittedLogEventsHandler == null) {
			deleteTransmittedLogEventsHandler = new CommandHandler() {
				public void execute(ReadOnlyCommandMetadata metadata,
						Object context) {
					if (canExecute(metadata, context)) {
						try {
							Logging4BBDatabaseManager.deleteTransmittedLogEventsFromDatabase();
						} catch (DatabaseOperationFailedException e) {
							// log
						}
					}
					return;
				}

				public boolean canExecute(ReadOnlyCommandMetadata metadata,
						Object context) {

					// TODO insert code to test if Command can be executed
					return true;
				}
			};
		}
		return deleteTransmittedLogEventsHandler;
	}

	public static Command deleteTransmittedLogEventsCommand() {
		if (deleteTransmittedLogEventsCmd == null) {
			deleteTransmittedLogEventsCommandMetaData = new CommandMetadata(deleteTransmittedLogEventsCommandID);
			deleteTransmittedLogEventsCommandMetaData.set(CommandMetadata.BITMAP_ICON,
					deleteTransmittedLogEventsCommandBitmapIcon);
			deleteTransmittedLogEventsCommandMetaData
					.set(CommandMetadata.ALIGNMENT, LEFT_ALIGNMENT);
			deleteTransmittedLogEventsCommandMetaData.setCommandCategories(new CategoryCollection(
					deleteTransmittedLogEventsCommandCategories));
			deleteTransmittedLogEventsCommandMetaData.setContextCategories(new CategoryCollection(
					deleteTransmittedLogEventsContextCategories));
			deleteTransmittedLogEventsCmd = new Command(deleteTransmittedLogEventsCommandHandler(),
					deleteTransmittedLogEventsCommandMetaData);
		}
		return deleteTransmittedLogEventsCmd;
	}

	public static CommandItem deleteTransmittedLogEventsCommandItem() {
		if (deleteTransmittedLogEventsCmdItem == null) {
			deleteTransmittedLogEventsCmdItem = new CommandItem(deleteTransmittedLogEventsCommand());
			deleteTransmittedLogEventsCmdItem.setIcon(deleteTransmittedLogEventsCommandImage());
			//deleteTransmittedLogEventsCmdItem.setText(ActionStringProviderFactory.deleteTransmittedRecords_Action());
			deleteTransmittedLogEventsCmdItem.setText(new StringProvider("deleteTransmittedLogEvents"));
		}
		return deleteTransmittedLogEventsCmdItem;
	}

	public static Bitmap deleteTransmittedLogEventsCommandBitmap() {
		if (deleteTransmittedLogEventsBitmap == null) {
			deleteTransmittedLogEventsBitmap = Bitmap.getBitmapResource(deleteTransmittedLogEventsCommandBitmapIcon);
		}
		return deleteTransmittedLogEventsBitmap;
	}

	public static Image deleteTransmittedLogEventsCommandImage() {
		if (deleteTransmittedLogEventsImage == null) {
			deleteTransmittedLogEventsImage = ImageFactory.createImage(deleteTransmittedLogEventsCommandBitmap());
		}
		return deleteTransmittedLogEventsImage;
	}
	
	// logEventFromRemoteClient
	private static Command logEventFromRemoteClientCmd;
	private static CommandItem logEventFromRemoteClientCmdItem;
	private static String logEventFromRemoteClientCommandID = "logEventFromRemoteClient";
	private static CommandMetadata logEventFromRemoteClientCommandMetaData;
	private static String[] logEventFromRemoteClientCommandCategories = { "logEventFromRemoteClient" };
	private static String[] logEventFromRemoteClientContextCategories = { Logging4BBApp.MY_APPLICATION_CONTEXT_CATEGORY };
	private static String logEventFromRemoteClientCommandBitmapIcon = "logEventFromRemoteClient_32.png";
	private static Image logEventFromRemoteClientImage;
	private static Bitmap logEventFromRemoteClientBitmap;
	private static CommandHandler logEventFromRemoteClientHandler;

	private static CommandHandler logEventFromRemoteClientCommandHandler() {
		if (logEventFromRemoteClientHandler == null) {
			logEventFromRemoteClientHandler = new CommandHandler() {
				public void execute(ReadOnlyCommandMetadata metadata,
						Object context) {
					if (canExecute(metadata, context)) {
						if (context instanceof Hashtable[]) {
							Hashtable[] events = (Hashtable[]) context;
							if (events.length == 0) {
								return;
							}
							if (Logging4BBApp.logToLogmonitor()) {
								if (events.length > 1) {
									Logging2LogMonitorSocketWorker.addTasks(events);
								} else {
									Logging2LogMonitorSocketWorker.addTask(new LogEventTask(events[0]));
								}
							}
							if (Logging4BBApp.logToWebsocket()) {
								// TODO
							}
							if (Logging4BBApp.logToDB()) {
								if (events.length > 1) {
									Logging2DatabaseWorker.addTasks(events);
								} else {
									Logging2DatabaseWorker.addTask(new LogEventTask(events[0]));
								}
							}
						}
					}
					return;
				}

				public boolean canExecute(ReadOnlyCommandMetadata metadata,
						Object context) {
//					if (context instanceof IExecutionContext) {
//						return ((IExecutionContext) context).canExecute();
//					}
					// TODO insert code to test if Command can be executed
					return true;
				}
			};
		}
		return logEventFromRemoteClientHandler;
	}

	public static Command logEventFromRemoteClientCommand() {
		if (logEventFromRemoteClientCmd == null) {
			logEventFromRemoteClientCommandMetaData = new CommandMetadata(logEventFromRemoteClientCommandID);
			logEventFromRemoteClientCommandMetaData.set(CommandMetadata.BITMAP_ICON,
					logEventFromRemoteClientCommandBitmapIcon);
			logEventFromRemoteClientCommandMetaData
					.set(CommandMetadata.ALIGNMENT, LEFT_ALIGNMENT);
			logEventFromRemoteClientCommandMetaData.setCommandCategories(new CategoryCollection(
					logEventFromRemoteClientCommandCategories));
			logEventFromRemoteClientCommandMetaData.setContextCategories(new CategoryCollection(
					logEventFromRemoteClientContextCategories));
			logEventFromRemoteClientCmd = new Command(logEventFromRemoteClientCommandHandler(),
					logEventFromRemoteClientCommandMetaData);
		}
		return logEventFromRemoteClientCmd;
	}

	public static CommandItem logEventFromRemoteClientCommandItem() {
		if (logEventFromRemoteClientCmdItem == null) {
			logEventFromRemoteClientCmdItem = new CommandItem(logEventFromRemoteClientCommand());
			logEventFromRemoteClientCmdItem.setIcon(logEventFromRemoteClientCommandImage());
			// logEventFromRemoteClientCmdItem.setText(ActionStringProviderFactory.logEventFromRemoteClient_Action());
			logEventFromRemoteClientCmdItem.setText(new StringProvider("logEventFromRemoteClient"));
		}
		return logEventFromRemoteClientCmdItem;
	}

	public static Bitmap logEventFromRemoteClientCommandBitmap() {
		if (logEventFromRemoteClientBitmap == null) {
			logEventFromRemoteClientBitmap = Bitmap.getBitmapResource(logEventFromRemoteClientCommandBitmapIcon);
		}
		return logEventFromRemoteClientBitmap;
	}

	public static Image logEventFromRemoteClientCommandImage() {
		if (logEventFromRemoteClientImage == null) {
			logEventFromRemoteClientImage = ImageFactory.createImage(logEventFromRemoteClientCommandBitmap());
		}
		return logEventFromRemoteClientImage;
	}
	
	//////////////////helper classes
	private static class CommandContextHashtable extends Hashtable {
		public String name;

		public CommandContextHashtable(String name) {
			this.name = name;
		}
		
		public String toString() {
			return name;
		}
	}

}
