/**
 *  Copyright (c) 2008...2012 Ekkehard Gentz (Rosenheim, Germany) and others.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *  
 *  Contributors:
 *          Ekkehard Gentz     - initial contribution 
 *     
 *   more info: see folder [project]/copyright
 */
package org.appbus.bb.app.logging4bb;

import java.util.Enumeration;

import javax.microedition.io.file.FileSystemListener;
import javax.microedition.io.file.FileSystemRegistry;

import net.rim.device.api.system.ApplicationManager;
import net.rim.device.api.system.Characters;
import net.rim.device.api.system.LED;
import net.rim.device.api.system.SystemListener;
import net.rim.device.api.system.WLANConnectionListener;
import net.rim.device.api.system.WLANInfo;
import net.rim.device.api.ui.UiApplication;
import net.rim.device.api.ui.component.Dialog;

import org.appbus.bb.app.logging4bb.api.ILogger;
import org.appbus.bb.app.logging4bb.db.Logging2DatabaseWorker;
import org.appbus.bb.app.logging4bb.db.Logging4BBDatabaseManager;
import org.appbus.bb.app.logging4bb.logger.Logger;
import org.appbus.bb.app.logging4bb.logger.LoggerFactory;
import org.appbus.bb.app.logging4bb.socket.Logging2LogMonitorSocketWorker;

public class Logging4BBApp extends UiApplication implements FileSystemListener, SystemListener {
	
	private static Logging4BBApp theApp;
	private static String SD_CARD_ROOT = "sdcard/";
	private static boolean startupCalled;
	private static boolean powerOffFired;
	
	// org.appbus.bb.app.logging4bb.Logging4BBApp.LogStoreID
	// 0x4e4a90ccba049a55L
	public static final ILogger logger = LoggerFactory.getLogger(Logging4BBApp.class);
	
	public static final long MY_OWN_STORE_ID = 0x4e4a90ccba049a55L;
	// TODO you can choose any name you want. this name will be used
	// as appname in SQLite DB or Lilith LogMonitor
	public static String MY_OWN_APP_NAME = "Logging4BBApp";	
	// TODO IMPORTANT: this name should match the name of your cod
	// and must also be used as LOGMANAGER_REMOTE_COMMAND_CONTEXT_CATEGORY from your client apps
	// please see example HelloLogging4bb
	public static String MY_APPLICATION_CONTEXT_CATEGORY = "logging4bb";
	
	// protocol and parameters
	private static String PROTOCOL = "socket://";
	private static String ON_DEVICE_SIDE = ";deviceside=false";
	// TODO you have to replace this
	private static String SEEKRET_CONNECTION = "??????????????????????????";
	
	// filename of the Database and location on SD card
	// TODO perhaps change this if you customize the logger
    public static String LOGGING4BB_DB_NAME = "LOGGING4BB.sqlite";
    public static String LOGGING4BB_DB_LOCATION = "/SDCard/databases/LOGGING4BB_DATA/";
	
	// PREEFERENCES
	private static int MY_VERSION = 1;
	private static int MY_DEFAULT_LOG_LEVEL = Logger.LOG_LEVEL_DEBUG;
	// if set to true: logging4bb is using the logger framework itself
	// if set to false: no logging happens
	private static boolean EAT_YOUR_OWN_DOGGFOOD = true;
	// TODO configure your host address
	// the IP where your Lilith is listening for LogEvents
	private static String HOST_ADDRESS = "127.0.0.1";
	// TODO port < 0 will be ignored
	// you need a server listening on that port
	// per ex. enmocs.org
	private static int PORT_TRANSMISSION = -1;
	// should not be changed !
	// Lilith is listening for the events on this port
	private static int PORT_LOGMONITOR = 10020;
	// for future versions
	private static int PORT_WEBSOCKET = -1;
	
	private static String SOCKET_SERVER_URL_FOR_TRANSMISSION;
	private static String SOCKET_SERVER_URL_FOR_LOGMONITOR;
	// we start with a small buffersize - used to cache logs before logging is initialized
	private static final int STARTUP_LOG_TASKS_BUFFER_SIZE = 500;
	// TODO this is the default maximum buffersize for log events
	// adjust it carefully to avoid problems if looping because of errors
	public static final int DEFAULT_LOG_TASKS_BUFFER_SIZE = 10000;
	private static int LOG_TAKS_BUFFER_SIZE = STARTUP_LOG_TASKS_BUFFER_SIZE;
	// as default we don't start in debug mode
	// debug mode means the LED is used to signal errors, warnings etc.
	private static boolean IS_DEBUG_MODE = false;
	// for future use: sending emails if ERROR or WARN was logged
	private static String ERROR_MAIL_TO = "";
	// TODO set this to true if you like to log to SQLite on your BB
	private static boolean LOG_TO_DB = false;
	// TODO as default we log to Lilith Logmonitor using socket connection
	private static boolean LOG_TO_LOGMONITOR = true;
	// for future versions
	private static boolean LOG_TO_WEBSOCKET = false;
	// only used if PORT_TRANSMISSION is set
	// TODO you can decide if transmissions happens always (true) or only inside WIFI (false)
	private static boolean TRANSMIT_ALWAYS = false;
	
	// transmission to server (only if PORT_TRANSMISSION > 0
	private boolean transmitRecordsToServerThreadRunning = false; 
	private boolean connectedToWlan = false;
	private long lastTransmitInsideWlan = Long.MIN_VALUE;
	private int restartIntervalInsideWlan = 30 * 60 * 1000; // ms for 30 minutes
	
    /**
     * Entry point for application
     * @param args Command line arguments (::led4debug:: used)
     */ 
    public static void main(String[] args) {
        // Create a new instance of the application and make the currently
        // running thread the application's event dispatch thread.
        theApp = new Logging4BBApp(); 
        // check EatYourOwnDogfood
        LoggerFactory.init(
        		MY_OWN_STORE_ID, 
        		MY_OWN_APP_NAME, 
        		MY_DEFAULT_LOG_LEVEL);
        theApp.logger.log("main", "early startup test");
        theApp.addSystemListener(theApp);
        theApp.addFileSystemListener(theApp);
        // test if system is running and SD Card is mounted
        if( rootExists4SD() && !ApplicationManager.getApplicationManager().inStartup() ) { 
        	delayedStartUp();
		} // else startup will be fired if power is up
        theApp.enterEventDispatcher();
    }
    

    /**
     * Creates a new Logging4BBApp object
     */
    public Logging4BBApp() {        
        // 
    }
	
    public boolean requestClose() {
    	logger.info("SHUT DOWN");
    	return true;
    }
    
    /**
    * Perform the start up work on a new Runnable using the 
    * invokeLater construct to ensure that it is executed
    * after the event thread has been created.
    * @see powerUp()
    *
    */
	public static void delayedStartUp(){
       	theApp.invokeLater( new Runnable() { 
            public void run() { 
          	  theApp.startMeUp();
            }
         } );
	}
	
	// will be started delayed after entering the event thread
	public void startMeUp(){
		startupCalled = true;
		
		// go to the back
		requestBackground();
		
		// init prefs from DB
		Logging4BBDatabaseManager.initPreferences();
		// TODO if not eatowndogfood: clear all caches of this apps logevents
		// if eatdogfood check logtodb, socket, websocket + clear cache if false
		
		// start the worker
		if (LOG_TO_DB) {
			Logging2DatabaseWorker.restartWorker(); //.startup();
			logger.debug("DB Worker started");
		} else {
			logger.debug("DB Worker not started");
		}
		if (LOG_TO_LOGMONITOR) {
			Logging2LogMonitorSocketWorker.restartWorker(); //.startup();
			logger.debug("LogMonitor Worker started");
		} else {
			logger.debug("LogMonitor Worker not started");
		}
		if (LOG_TO_WEBSOCKET) {
			// TODO start websocket task worker
		}
		
		// start Listener 4 WLAN ?
		if (HOST_ADDRESS != null && HOST_ADDRESS.length() > 0 && PORT_TRANSMISSION > 0) {
			// if log events are stored into SQLite, then
			// if inside a WLAN we send the events to a central server
			startListener4WLAN();
			logger.debug("WLAN Listener started");
		} else {
			logger.debug("WLAN Listener not started");
		}
		
		// now we can register global commands,
		// because the app is listening for incoming tasks
		CommandFactory.registerRemoteCommands();
	}
	
	public void notifyForeground(final String message) {
		if (message != null) {
			theApp.requestForeground();
			UiApplication.getUiApplication().invokeLater(new Runnable(){
                public void run(){
                	// TODO insert Message to notify instead of Dialog
                	Dialog.inform(message);
                	theApp.requestBackground();
                } 
            }); 
		}
	}
	
	/**
	 * in special cases where you want to exit the app
	 * you can notify the user and shut dwon
	 * 
	 * @param message
	 */
    public void shutMeDown(final String message){
    	if (message != null) {		
    		theApp.requestForeground();
    		UiApplication.getUiApplication().invokeLater(new Runnable(){
                public void run(){
                	if (IS_DEBUG_MODE) {
                		try {
    						LED.setState(LED.STATE_OFF);
    					} catch (Throwable e) {
    						// ignore
    					}
					}
                	Dialog.alert(message);
                } 
            }); 
    	} 
    	if (LOG_TO_DB) {
    		Logging2DatabaseWorker.shutdown();
		}
    	if (LOG_TO_LOGMONITOR) {
			Logging2LogMonitorSocketWorker.shutdown();
		}
    	// TODO WebSocket
       	System.exit(0);
    }
     
    
    public static Logging4BBApp getInstance(){
    	return theApp;
    }
   
    
    // TRANSMISSION TO SERVER
    public void setTransmitLogEventsToServerThreadRunning(boolean running) {
		transmitRecordsToServerThreadRunning = running;
	}

	public boolean isTransmitLogEventsToServerThreadRunning() {
		return transmitRecordsToServerThreadRunning;
	}
	
	public void setConnectedToWlan(boolean connectedToWlan) {
		this.connectedToWlan = connectedToWlan;
	}


	/**
	 * returns true if connected to WLAN
	 * OR if always-transmit was set (per ex. backdoor altSEND)
	 * 
	 * @return
	 */
	public boolean isConnectedToWlan() {
		if (transmitAlways()) {
			return true;
		} else {
			return connectedToWlan;
		}
	}
	
	/**
	 * sets the time of last transmission if still in WLAN
	 * 
	 * @param lastTransmitInsideWlan
	 */
	public void setLastTransmitInsideWlan(long lastTransmitInsideWlan) {
		if (connectedToWlan) {
			this.lastTransmitInsideWlan = lastTransmitInsideWlan;
		} else {
			this.lastTransmitInsideWlan = Long.MIN_VALUE;
		}
	}


	public void startTransmissionIfWaitingTooLongInsideWlan() {
		if (lastTransmitInsideWlan != Long.MIN_VALUE
				&& (System.currentTimeMillis()-lastTransmitInsideWlan) > restartIntervalInsideWlan ) {
			lastTransmitInsideWlan = Long.MIN_VALUE;
			startTransmissionIfNotRunning();
		}
	}
	
	private void startTransmissionIfNotRunning(){
		if (!transmitRecordsToServerThreadRunning) {
			// START
    		CommandFactory.startTransmissionToServerCommand().execute(null);
		}
	}

	public static int getVersion(){
		return MY_VERSION;
	}
	
	public static boolean eatYourOwnDogfood(){
		return EAT_YOUR_OWN_DOGGFOOD;
	}
	
	public static void setEatYourOwnDogfood(boolean b){
		EAT_YOUR_OWN_DOGGFOOD = b;
	}
	
	public static int getDefaultLogLevel(){
		return MY_DEFAULT_LOG_LEVEL;
	}
	
	public static void setDefaultLogLevel(int level){
		if (level >= 0 && level < 4) {
			MY_DEFAULT_LOG_LEVEL = level;
			LoggerFactory.setDefaultLogLevel(level);
		}
	}

	public static String getHostAddress(){
		return HOST_ADDRESS;
	}

	public static void setHostAddress(String hostAddress) {
		HOST_ADDRESS = hostAddress;
	}

	public static int getPortLogmonitor(){
		return PORT_LOGMONITOR;
	}
	
	public static void setPortLogmonitor(int p){
		PORT_LOGMONITOR = p;
	}
	
	public static int getPortTransmission(){
		return PORT_TRANSMISSION;
	}
	
	public static void setPortTransmission(int p){
		PORT_TRANSMISSION = p;
	}
	
	public static boolean transmitAlways(){
		return TRANSMIT_ALWAYS;
	}
	
	public static void setTransmitAlways(boolean transmitAlways) {
		TRANSMIT_ALWAYS = transmitAlways;
	}
	
	public static int getPortWebsocket(){
		return PORT_WEBSOCKET;
	}
	
	public static void setPortWebsocket(int p){
		PORT_WEBSOCKET = p;
	}
	
	public static boolean logToDB() {
		return LOG_TO_DB;
	}
	
	public static void setLogToDB(boolean b){
		LOG_TO_DB = b;
	}
	
	public static boolean logToLogmonitor() {
		return LOG_TO_LOGMONITOR;
	}
	
	public static void setLogToLogmonitor(boolean b){
		LOG_TO_LOGMONITOR = b;
	}
	
	public static boolean logToWebsocket() {
		return LOG_TO_WEBSOCKET;
	}
	
	public static void setLogToWebsocket(boolean b) {
		LOG_TO_WEBSOCKET = b;
	}

	public static void setLogTasksBufferSize(int logTasksBufferSize) {
		LOG_TAKS_BUFFER_SIZE = logTasksBufferSize;
		Logging2DatabaseWorker.ensureCapacity(logTasksBufferSize);
	}

	public static int getLogTasksBufferSize() {
		return LOG_TAKS_BUFFER_SIZE;
	}

	public static boolean isDebugMode() {
		return IS_DEBUG_MODE;
	}
	
	public static void setDebugMode(boolean isDebugMode) {
		IS_DEBUG_MODE = isDebugMode;
	}
	
	public static String getErrorMailTo(){
		return ERROR_MAIL_TO;
	}
	
	public static void setErrorMailTo(String s) {
		ERROR_MAIL_TO = s;
	}

	public static String getSocketServerUrl4Transmission(){
		if (SOCKET_SERVER_URL_FOR_TRANSMISSION == null) {
			// init URL
			StringBuffer sb = new StringBuffer();
	        sb.append(PROTOCOL);
	        sb.append(HOST_ADDRESS);
	        sb.append(Characters.COLON);
	        sb.append(PORT_TRANSMISSION);
			sb.append(ON_DEVICE_SIDE);
			sb.append(SEEKRET_CONNECTION);
			SOCKET_SERVER_URL_FOR_TRANSMISSION = sb.toString();
		}
		return SOCKET_SERVER_URL_FOR_TRANSMISSION;
    }
	
	public static String getSocketServerUrl4LogMonitor(){
		if (SOCKET_SERVER_URL_FOR_LOGMONITOR == null) {
			// init URL
			StringBuffer sb = new StringBuffer();
	        sb.append(PROTOCOL);
	        sb.append(HOST_ADDRESS);
	        sb.append(Characters.COLON);
	        sb.append(PORT_LOGMONITOR);
			sb.append(ON_DEVICE_SIDE);
			sb.append(SEEKRET_CONNECTION);
			SOCKET_SERVER_URL_FOR_LOGMONITOR = sb.toString();
		}
		return SOCKET_SERVER_URL_FOR_LOGMONITOR;
    }
    
    
    /////// SYSTEM LISTENER

	

	// maybe stop logging and notify user if battery low
	// and restart logging if battery good again
	public void batteryLow() {}
	public void batteryGood() {}
	public void batteryStatusChange(int status) {}

	public void powerOff() {
		powerOffFired = true;
	}


	public void powerUp() {
		// handle the special case of powerUp after powerOff
		if (powerOffFired) {	
			powerOffFired = false;
			// nothing special with logging to DB - 
			// runs well after powerOff / powerUp
			if (LOG_TO_LOGMONITOR) {
				// the socket connection to Lilith doesn't survive
				// because directly after powerUp() not all TransportTypes are ready
				// so we restart
				Logging2LogMonitorSocketWorker.restartWorker();
				logger.debug("Socket LogMonitor restarted after powerOff");
			}
			logger.debug("powerUp after powerOff");
			return;
		}
		// test at powerUp if initialization already done
		boolean rootExists = rootExists4SD();
		if (rootExists && !startupCalled) {
			logger.debug("powerUp, SD is ready - startMeUp now");
			startMeUp();
			logger.debug("startMeUp from PowerUp finished");
		} else if (!rootExists) {
			logger.debug("powerUp, not started, SD missing");
		} else {
			logger.debug("powerUp, but I'm already started");
		}
	}
	
	//////////////////////////////////////////////

	/**
	 * File System Listener
	 * 
	 * its important to watch the state of SD Root
	 * if BlackBerry is used as USB Drive to copy
	 * fotos, music etc then the SD Card was unmounted
	 * for BlackBerry Apps and SQLite cannot be used
	 * we have to wait until SD Root comes back
	 * 
	 * @param state
	 * @param rootName
	 */
	public void rootChanged(int state, String rootName) {
		switch (state) {
		case FileSystemListener.ROOT_ADDED:
			if (rootName.equalsIgnoreCase(SD_CARD_ROOT)) {
				logger.debug("SD Card added");
				if (!startupCalled) {
					startMeUp();
					logger.debug("startMeUp from SD added finished");
				} else {
					Logging4BBDatabaseManager.resumeDatabase();
					// notifyForeground("SD added");
				}
			}
			break;
		case FileSystemListener.ROOT_REMOVED:
			if (rootName.equalsIgnoreCase(SD_CARD_ROOT)) {
				logger.debug("SD Card removed");
				// stop database + pause tracker
				Logging4BBDatabaseManager.pauseDatabase();
				// notifyForeground("SD removed");
			}
			break;
		default:
			// nothing
			break;
		}
		
	}
	
	// TODO register status command to ask the tracker about state
	// if sd card is there etc
	private static boolean rootExists4SD(){
        Enumeration roots;
        try {
			roots = FileSystemRegistry.listRoots();
		} catch (SecurityException e) {
			// logger.warn("SecurityException trying to ask FileSystemRegistry for Roots: ", e);
			return false;
		} catch (Throwable e) {
			// logger.error("Error trying to ask FileSystemRegistry for Roots: ", e);
			return false;
		}
        while (roots.hasMoreElements()) {
			String rootName = (String) roots.nextElement();
			if (rootName.equalsIgnoreCase(SD_CARD_ROOT)) {
				return true;
			}
		}
        return false;
	}
	
	/**
	 * WLAN Listener
	 * checks current state of WLAN connection
	 * if connected, starts transmission of data
	 * then adds Listener
	 * Listener sets WLAN connection to true or false
	 * and restarts Data transmission if connected and nothing runs
	 */
	public void startListener4WLAN(){
		if (WLANInfo.getWLANState() == WLANInfo.WLAN_STATE_CONNECTED) {
			// TODO Log.info("network Connected at startup");
        	connectedToWlan = true;
        	startTransmissionIfNotRunning();
		}
		WLANInfo.addListener(new WLANConnectionListener() {
            public void networkConnected() {
                // TODO Log.info("network Connected");
            	connectedToWlan = true;
            	startTransmissionIfNotRunning();
            }
            public void networkDisconnected(int i) {
                // TODO Log.info("network disconnected");
            	// this will cause the transmission to stop after next iteration
            	connectedToWlan = false;
            	// reset timestamp of last transmission inside WLAN
            	lastTransmitInsideWlan = Long.MIN_VALUE;
            }
        });
	}

}
