/**
 *  Copyright (c) 2008...2012 Ekkehard Gentz (Rosenheim, Germany) and others.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *  
 *  Contributors:
 *          Ekkehard Gentz     - initial contribution 
 *          				   - with ideas from BlackBerry Socket Demo
 *     
 *   more info: see folder [project]/copyright
 */

package org.appbus.bb.app.logging4bb.db;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;

import javax.microedition.io.StreamConnection;

import net.rim.device.api.io.transport.ConnectionDescriptor;
import net.rim.device.api.io.transport.ConnectionFactory;
import net.rim.device.api.io.transport.TransportInfo;
import net.rim.device.api.system.Characters;
import net.rim.device.api.system.DeviceInfo;
import net.rim.device.api.util.LongEnumeration;
import net.rim.device.api.util.LongHashtable;
import net.rim.device.api.util.LongVector;

import org.appbus.bb.app.logging4bb.CommandFactory;
import org.appbus.bb.app.logging4bb.Logging4BBApp;
import org.appbus.bb.app.logging4bb.util.HexUtil;

/**
 * A thread class to handle communication with the server component.
 */
public class TransmitLogEventsToServer extends Thread {  
	
	// Server and Client APP are always paired and belong to a fix Port
	// org.ekkescorner.tcpsocket.server.logging4bb
	// 0x514a396d9d34f648L
	private static String TCP_SOCKET_SERVER_CONNECTION_ID = "514a396d9d34f648";
    // org.ekkescorner.tcpsocket.client.logging4bb
	// 0xeb0cddc56c56f8dbL
	private static String TCP_SOCKET_CLIENT_CONNECTION_ID = "eb0cddc56c56f8db";
	
	// we have only one TASK at the moment
	// org.ekkescorner.tcpsocket.client.logging4bb.task.transmit.records
	// 0xf916d9cae3e727d5L
	private static String TCP_SOCKET_TASK_ID_GPS_RECORDS_TRANSMIT_TO_SERVER = "f916d9cae3e727d5";
	
	// org.ekkescorner.tcpsocket.blackberryfixlength.internal.BlackBerryFixLengthWorker.wrongid
	// 0xfd21d5e7ee6de0efL
	private static String TCP_SOCKET_SERVER_WRONG_ID = "fd21d5e7ee6de0ef";
	// org.ekkescorner.tcpsocket.blackberryfixlength.internal.BlackBerryFixLengthWorker.wronglength
	// 0xf77b13f2d92f3960L
	private static String TCP_SOCKET_SERVER_WRONG_LENGTH = "f77b13f2d92f3960";
	// org.ekkescorner.tcpsocket.blackberryfixlength.internal.BlackBerryFixLengthWorker.datareaderror
	// 0x11f71efe1ca7dcbcL
	private static String TCP_SOCKET_SERVER_DATA_READ_ERROR = "11f71efe1ca7dcbc";
	// org.ekkescorner.tcpsocket.blackberryfixlength.internal.BlackBerryFixLengthWorker.datastructureerror
	// 0x35d7067e81e2000cL
	private static String TCP_SOCKET_SERVER_DATA_STRUCTURE_ERROR = "35d7067e81e2000c";
	// org.ekkescorner.tcpsocket.blackberryfixlength.internal.BlackBerryFixLengthWorker.dataprocesserror
	// 0x434c6905beb15bfdL
	private static String TCP_SOCKET_SERVER_DATA_PROCESS_ERROR = "434c6905beb15bfd";
	// org.ekkescorner.tcpsocket.blackberryfixlength.internal.BlackBerryFixLengthWorker.devicenotregistered
	// 0x298d6ea0b346b861L
	private static String TCP_SOCKET_SERVER_DEVICE_NOT_REGISTERED = "298d6ea0b346b861";
	// org.ekkescorner.tcpsocket.blackberryfixlength.internal.BlackBerryFixLengthWorker.unknowntask
	// 0x8d4185940ebc43eaL
	private static String TCP_SOCKET_SERVER_UNKNOWN_TASK = "8d4185940ebc43ea";
	// org.ekkescorner.tcpsocket.blackberryfixlength.internal.BlackBerryFixLengthWorker.transmissionnumbererror
	// 0x8e8ba72556405818L
	private static String TCP_SOCKET_SERVER_TRANSMISSION_NUMBER_ERROR = "8e8ba72556405818";
	// org.ekkescorner.tcpsocket.blackberryfixlength.internal.BlackBerryFixLengthWorker.wrongproperties
	// 0x1a4fe8f0d809bf06L
	private static String TCP_SOCKET_SERVER_WRONG_PROPERTIES = "1a4fe8f0d809bf06";
	
	// position 0 from char array of properties
	private static final int PROPERTY_CONNECTION_TYPE_POSITION = 0;
	// possible valies of this propertoie
	private static final char PROPERTY_CONNECTION_TYPE_IS_SEND_DATA = '1';

	private static String EMPTY_PROPERTIES = "0000000000000000";
	
	// demo request properties 1 = send data to server 3 = get data
	private static String DEMO_REQUEST_SEND_DATA_PROPERTIES = "1000000000000000";
	// demo response properties 1 = send data to server
	private static String DEMO_RESPONSE_SEND_DATA_PROPERTIES = "0100000000000000";
	
	private InputStream in;
    private OutputStreamWriter out;  
    private StreamConnection connection;
    private StringBuffer receiveBuffer;
    private String transmission;
    private StringBuffer sb;
    private byte[] bytes;
    private long lengthOfData2GetFromServer;
    
    private LongHashtable results;
    private String deviceId;
    
    // Constructor
    public TransmitLogEventsToServer(LongHashtable results) {

        this.results = results;
    }
    
    /**
     * use-case SEND_DATA
     * 
     * sends the request to the server to send some data
     * DEVICE, TASK, TRANSMISSION, PROPERTIES, LENGTHINFO
     * 
     * @param client
     * @param device
     * @param task
     * @param transmission
     * @param properties
     * @param dataLength
     * @throws IOException
     */
    private void sendRequest2SendDataToServer(String client,
    		String device,
    		String task, 
    		String transmission, 
    		String properties, 
    		long dataLength
    		) throws IOException{
        String dataLengthAsHex = HexUtil.fromLong(dataLength);
    	StringBuffer sendBuffer = new StringBuffer();
    	sendBuffer.append(client);
    	sendBuffer.append(device);
    	sendBuffer.append(task);
    	sendBuffer.append(transmission);
    	sendBuffer.append(properties);
    	sendBuffer.append(dataLengthAsHex);
    	
    	out.write(sendBuffer.toString(), 0, sendBuffer.length());
    	out.flush();
    }
    
    
    /**
     * tries to read the next 16 Bytes
     * if there are less then 16 read then returns false
     * 
     * @return
     * @throws IOException
     */
    private boolean getNext16Bytes() throws IOException{
    	receiveBuffer = new StringBuffer(16);
    	bytes = new byte[16];
        int totalBytesReadSoFar = 0;
        int bytesReadThisIteration = 0;
        int numBytesToRead = 16;     
        while (totalBytesReadSoFar < numBytesToRead) {
        	 //Loops until it receives the necessary amount of bytes, and stores the accumulated bytes
             // in receiveBuffer  
             // we have to limit the maximum to 16 bytes
        	 // all these Strings are filled with HEX Values 0...9 a...f
         		bytesReadThisIteration = in.read(bytes, totalBytesReadSoFar, 16-totalBytesReadSoFar);
         		if (bytesReadThisIteration == -1) {
         			// reached the end
         			// go on -we check the results soon
    					break;
    				}
                 receiveBuffer.append(new String(bytes, totalBytesReadSoFar, bytesReadThisIteration));
                 totalBytesReadSoFar += bytesReadThisIteration;
    	 }     
         return (totalBytesReadSoFar > 0 && totalBytesReadSoFar == numBytesToRead);
    }
    
    private void logIntoDB(final String message){
    	Logging2DatabaseWorker.addTask(message);
    }
    
    /**
     * checks the returncode for ERROR messages
     * and displays the message in statusfield of main screen
     * 
     * @param receivedAnswer
     */
    private void displayErrorMessage(String receivedAnswer){
        if (receivedAnswer.equals(TCP_SOCKET_SERVER_DATA_READ_ERROR)) {
        	logIntoDB("DATA READ ERROR  ...END");
        	return;
		} 
        
        if (receivedAnswer.equals(TCP_SOCKET_SERVER_WRONG_ID)) {
        	logIntoDB("Server: WRONG ID  ...END");
        	return;
		} 
        
        if (receivedAnswer.equals(TCP_SOCKET_SERVER_WRONG_LENGTH)) {
        	logIntoDB("Server: WRONG LENGTH received  ...END");
        	return;
		} 
        
        if (receivedAnswer.equals(TCP_SOCKET_SERVER_DEVICE_NOT_REGISTERED)) {
        	logIntoDB("Server: Device not registered for this service  ...END");
        	return;
		} 
        
        if (receivedAnswer.equals(TCP_SOCKET_SERVER_TRANSMISSION_NUMBER_ERROR)) {
        	logIntoDB("Server: Transmission Number Error  ...END");
        	return;
		} 
        
        if (receivedAnswer.equals(TCP_SOCKET_SERVER_DATA_STRUCTURE_ERROR)) {
        	logIntoDB("Server: Data Structure Error  ...END");
        	return;
		} 
        
        if (receivedAnswer.equals(TCP_SOCKET_SERVER_UNKNOWN_TASK)) {
        	logIntoDB("Server: Unknown Task ID ...END");
        	return;
		}
        
        if (receivedAnswer.equals(TCP_SOCKET_SERVER_WRONG_PROPERTIES)) {
        	logIntoDB("Server: wrong Properties ...END");
        	return;
		}
        
        // now we have to encode UTF-8
        String uuups = null;
		try {
			uuups = new String(bytes,0,Math.max(16, bytes.length),"UTF-8");
		} catch (UnsupportedEncodingException e) {
			// 
		}
		logIntoDB("UUUPS: unexpected response value: "+uuups);
    	return;
    }
    
    /**
     * use-case SEND_DATA
     * 
     * checks response from server
     * TASK, TRANSMISSION, PROPERTY, LENGTHINFO
     * @return
     */
    private boolean getResponse2SendDataFromServer(){
        // W A I T   F O R   R E S P O N S E   C O D E
    	boolean got16Bytes;
    	String received = null;
		try {
			// get the task ID || error ID if failed at server side
			got16Bytes = getNext16Bytes();
			received = receiveBuffer.toString();
			if (got16Bytes) {
				if (!received.equals(TCP_SOCKET_TASK_ID_GPS_RECORDS_TRANSMIT_TO_SERVER)) {
					displayErrorMessage(received);
					return false;
				}
			}
			// get the transmission number
			got16Bytes = getNext16Bytes();
			received = receiveBuffer.toString();
			if (got16Bytes) {
				if (!received.equals(transmission)) {
					logIntoDB("not the same TRANSMISSION as sent to server");
					displayErrorMessage(received);
					return false;
				}
			}
			// get the properties
			got16Bytes = getNext16Bytes();
			received = receiveBuffer.toString();
			if (got16Bytes) {
				if (!received.equals(DEMO_RESPONSE_SEND_DATA_PROPERTIES)) {
					logIntoDB("not the expected RESPONSE PROPERTIES from Server");
					displayErrorMessage(received);
					return false;
				}
			}
			// get the length info
			got16Bytes = getNext16Bytes();
			received = receiveBuffer.toString();
			if (got16Bytes) {
				if (!received.equals(EMPTY_PROPERTIES)) {
					logIntoDB("not the expected LENGTH INFO from Server");
					displayErrorMessage(received);
					return false;
				}
			}
			// response OK
		} catch (IOException e) {
			logIntoDB("IOException while reading next 16 bytes: "+e.getMessage());
			return false;
		}
        // we'll catch this if connection was made thru BES MDS
        // if direct Connection then we get immediately connection refused much earlier
        if (!got16Bytes) {
        	logIntoDB("unsuccessfully END - need 16 bytes, but read only "+received);
        	return false;
		}
        return true;
    }
    
     
    /**
     * use-case SEND_DATA
     * 
     * if all went well at the end the server sends his
     * CONNECTION_ID back
     * 
     * @return
     */
    private boolean getAcknowledgeFromServer(){
    	boolean success;
		try {
			success = getNext16Bytes();
		} catch (IOException e) {
			logIntoDB("IOException while reading next 16 bytes: "+e.getMessage());
			return false;
		}
        // we'll catch this if connection was made thru BES MDS
        // if direct Connection then we get immediately connection refused much earlier
        if (!success) {
        	logIntoDB("unsuccessfully END - need 16 bytes, but read only "+receiveBuffer.toString());
        	return false;
		}
        // no encoding needed - only 0..1a..f
        if (receiveBuffer.toString().equals(TCP_SOCKET_SERVER_CONNECTION_ID)) {
        	// TODO logIntoDB(bean.getAktionsText()+"an Server gesendet...", bean.getAktion());
        	return true;
		} 
        displayErrorMessage(receiveBuffer.toString());
        return false;
    }
    
    /**
     * use-case SEND_DATA
     * 
     * sends the data together with
     * TASK, TRANSMISSION, PROPERTIES, LENGTHINFO
     * to the server
     * 
     * @param task
     * @param transmission
     * @param properties
     * @param dataLenth
     * @param data
     * @throws IOException
     */
    private void sendDataToServer(
    		String task, 
    		String transmission, 
    		String properties, 
    		long dataLenth,
    		String data
    		) throws IOException {
        String dataLengthAsHex = HexUtil.fromLong(dataLenth);
    	StringBuffer sendBuffer = new StringBuffer();
    	sendBuffer.append(task);
    	sendBuffer.append(transmission);
    	sendBuffer.append(properties);
    	sendBuffer.append(dataLengthAsHex);
    	sendBuffer.append(data);
        out.write(sendBuffer.toString(), 0, sendBuffer.length());
        out.flush();
    }
    
    
    /**
     * cleaning all what's possible
     * closing in- and out - Streams
     * closing the connection
     * 
     */
    private void cleanUp(){
        Logging4BBApp.getInstance().setTransmitLogEventsToServerThreadRunning(false); 
        try {               
            if (in != null) {
            	in.close();     
			}            
        }
        catch(IOException ioe) {       
        	// nothing
        }
        try {       
            if (out != null) {
            	out.close();
			}	             
        }
        catch(IOException ioe) {   
        	// nothing
        }
        try {               
            if (connection != null) {
            	connection.close();
			} 	
        }
        catch(IOException ioe) {               
        	// nothing
        }
    }
    
    /**
     * constructs the URL, opens the Connection
     * opens input- and output streams
     * 
     * @return
     */
    private boolean initializeConnectionInAndOut(){
		String url = Logging4BBApp.getSocketServerUrl4Transmission();
    	try {
    		ConnectionFactory cf = new ConnectionFactory();
    		cf.setPreferredTransportTypes(new int[]{
    				TransportInfo.TRANSPORT_TCP_WIFI,
    				TransportInfo.TRANSPORT_MDS,
    				TransportInfo.TRANSPORT_BIS_B,
    				TransportInfo.TRANSPORT_TCP_CELLULAR,
    				TransportInfo.TRANSPORT_WAP2});
    		ConnectionDescriptor cd = cf.getConnection(url);
    		if (cd == null) {
    			logIntoDB("error open: no Connection descriptor "+url);
    			cleanUp();
            	return false;
			}
    		connection = (StreamConnection) cd.getConnection();
		} catch (Exception e) {
			logIntoDB("error open: "+url+" "+e.getMessage());
			cleanUp();
        	return false;
		}
        if (connection == null) {
        	logIntoDB("Connection NULL: cannot open"+url);
        	cleanUp();
        	return false;
		}
        //logIntoDB("START connection...");
        try {
			in = connection.openInputStream();
		} catch (IOException e) {
			logIntoDB("IOException open InputStream: "+e.getMessage());
			cleanUp();
			return false;
		}
                    
        try {
			out = new OutputStreamWriter(connection.openOutputStream(), "UTF-8");
		} catch (UnsupportedEncodingException e) {
			logIntoDB("ensupported encoding: "+e.getMessage());
			// fall back to default
			try {
				out = new OutputStreamWriter(connection.openOutputStream());
			} catch (IOException e2) {
				logIntoDB("IOException opening Output Stream" + e2.getMessage());
				cleanUp();
				return false;
			}
		}  catch (IOException e) {
			logIntoDB("IOException open encoded OutputStream: "+e.getMessage());
			cleanUp();
			return false;
		}  
		return true;
    }

   /**
    * Implementation of Thread.
    */
    public void run() {        
        connection = null;    
        // default if connected to simulator 0x000000002100000A
        deviceId = HexUtil.fromLong(DeviceInfo.getDeviceId());   
        ///
        LongEnumeration enum = results.keys();
		int max = 10;
		int count = 0;
		StringBuffer transmissionBuffer = new StringBuffer();
		LongVector recordIds = new LongVector(max);
		boolean doItAgain = true;
		while (enum.hasMoreElements()) {
			long key = enum.nextElement();
			count++;
			if (count > max) {
				// transmit records
				boolean success = transmitSomeRecord(transmissionBuffer.toString());
				if (!success) {
					// now we don't update the records as transmitted
					// perhaps server already has inserted some
					// this doesn't matter because duplicates will
					// be ignored next time transmitting
					doItAgain = false;
					break;
				}
				// update buffered keys
				try {
					Logging4BBDatabaseManager.updateLogEventsTransmitted(recordIds);
				} catch (DatabaseOperationFailedException e) {
					// TODO LOG
					doItAgain = false;
					break;
				}
				// reset buffer
				transmissionBuffer = new StringBuffer();
				recordIds = new LongVector(max);
				count =1;
				// check if still in WLAN
				if (!Logging4BBApp.getInstance().isConnectedToWlan()) {
					// TODO LOG
					doItAgain = false;
					break;
				}
			}
			// add record to transmissionBuffer
			// if not first record we have to add one delimiter first
			if (count > 1) {
				transmissionBuffer.append(Characters.COMMA);
			}
			transmissionBuffer.append(((StringBuffer)results.get(key)).toString());
			// collect keys
			recordIds.addElement(key);
		}
		// all went well AND some records collected but not transmitted ?
		if (doItAgain && recordIds.size() > 0) {
			// transmit records
			boolean success = transmitSomeRecord(transmissionBuffer.toString());
			if (success) {
				// update buffered keys
				try {
					Logging4BBDatabaseManager.updateLogEventsTransmitted(recordIds);
				} catch (DatabaseOperationFailedException e) {
					// TODO LOG
					doItAgain = false;
				}
			} else {
				doItAgain = false;
			}
		}
        // cleanUp();
		Logging4BBApp.getInstance().setTransmitLogEventsToServerThreadRunning(false);
		// start the process again async: select next 100 records and process
		// only start if in WLAN
		if (doItAgain && Logging4BBApp.getInstance().isConnectedToWlan()) {
			CommandFactory.startTransmissionToServerCommand().execute(null);
		}
    }
    
    // uses the default way to transmit one action per connection
    // TODO at server side another workflow to transmit using same connection
    private boolean transmitSomeRecord(String transmissionBuffer){
    	if (!initializeConnectionInAndOut()) {
			return false;
		}
    	
    	// transmission number: unique number counting the transmissions
        // could be that server gets last time all data, but client didn't get
        // the confirmation because of an error
        // so if the client wants to resend the same transmission
        // the server cann notice that the data already was received
        transmission = HexUtil.fromLong(42);
		long utfLength = 0;
		try {
			// SEND VALID_DATA with no Umlauts and some rows of data
			utfLength = transmissionBuffer.getBytes("UTF-8").length;
			sendRequest2SendDataToServer(
					TCP_SOCKET_CLIENT_CONNECTION_ID, // client id OK
					deviceId, // device ID OK
					TCP_SOCKET_TASK_ID_GPS_RECORDS_TRANSMIT_TO_SERVER, // TASK ID OK
					transmission, // Transmission ID OK
					DEMO_REQUEST_SEND_DATA_PROPERTIES, // properties OK
					utfLength // correct length of data
					); 
			if (getResponse2SendDataFromServer()) {
				sendDataToServer(
						TCP_SOCKET_TASK_ID_GPS_RECORDS_TRANSMIT_TO_SERVER, // TASK ID OK
			    		transmission, // Transmission ID OK
			    		DEMO_REQUEST_SEND_DATA_PROPERTIES, // properties OK
			    		utfLength, // correct length of data
			    		transmissionBuffer // correct DATA
			    		); 
			    getAcknowledgeFromServer();
			}
		} catch (UnsupportedEncodingException e) {
			// TODO LOG
			cleanUp();
			return false;
		} catch (IOException e) {
			// TODO LOG
			logIntoDB("IO Exception: transmit one record: "+e.getMessage());
			cleanUp();
			return false;
		}
		cleanUp();
    	return true;
    }
}
