/**
 *  Copyright (c) 2008...2012 Ekkehard Gentz (Rosenheim, Germany) and others.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *  
 *  Contributors:
 *          Ekkehard Gentz     - initial contribution 
 *     
 *   more info: see folder [project]/copyright
 */
package org.appbus.bb.app.logging4bb.logger;

import java.util.Vector;

import net.rim.device.api.system.Characters;

import org.appbus.bb.app.logging4bb.Logging4BBApp;
import org.appbus.bb.app.logging4bb.api.ILogger;
import org.appbus.bb.app.logging4bb.db.Logging2DatabaseWorker;
import org.appbus.bb.app.logging4bb.model.LogEventTask;
import org.appbus.bb.app.logging4bb.socket.Logging2LogMonitorSocketWorker;

/**
 * 
 * local Logger Implementation to log events inside this app
 * see folder /remoteClients for remote Clients -
 * Logger implementations.
 * there's a FallbackLogger if no logging4bb app is running
 * and a default RemoteLogger implementation
 * calling this app using Remote Commands
 * 
 * @author ekkehard gentz (ekke), rosenheim, germany
 *
 */
public class Logger implements ILogger{
	


	
    // its a good idea to have a separate Logger for each Class
	private Class clazz;
	
	// private static SocketLoggerThread loggingThread;
	/**
	 * initializes the Logger for Class clazz
	 * 
	 * @param clazz
	 */
	public Logger (Class clazz){
		this.clazz = clazz;
	}
	
	private void logEvent(String context, String message, int level, Throwable exception){
		if (LoggerFactory.DEFAULT_LOG_LEVEL > level) {
			return;
		}
		if (Logging4BBApp.eatYourOwnDogfood()) {
			LogEventTask logEvent = new LogEventTask(
					LoggerFactory.LOG_STORE_ID, 
					LoggerFactory.APP_NAME, 
					clazz.getName());
			logEvent.setContext(context);
			logEvent.setMessage(message);
			logEvent.setLevel(level);
			logEvent.setEventTime(System.currentTimeMillis());
			if (exception != null) {
				logEvent.setException(exception);
			}
			if (Logging4BBApp.logToDB()) {
				Logging2DatabaseWorker.addTask(logEvent);
			}
			if (Logging4BBApp.logToLogmonitor()) {
				Logging2LogMonitorSocketWorker.addTask(logEvent);
			}
			if (Logging4BBApp.logToWebsocket()) {
				// TODO
			}
		}
		// fallback if debugging
		if (context != null && context.length() > 0) {
			message = context + Characters.COLON  + message;
		}
		System.out.println(
				logLevelTexts[level]
				+Characters.SPACE                       
				+LoggerFactory.APP_NAME
				+Characters.COLON+Characters.SPACE
				+message);
		if (exception != null) {
			System.out.println(
					logLevelTexts[level]
	 				+Characters.SPACE                       
	 				+LoggerFactory.APP_NAME
	 				+Characters.COLON+Characters.SPACE
					+"got Exception: "
					+exception.getMessage());
			exception.printStackTrace();
		}
	}

	public void debug(String logMessage) {
		logEvent(null, logMessage, LOG_LEVEL_DEBUG, null);
	}
	
	public void debug(String logMessage, Throwable exception) {
		logEvent(null, logMessage, LOG_LEVEL_DEBUG, exception);
	}

	public void debug(String context, String logMessage) {
		logEvent(context, logMessage, LOG_LEVEL_DEBUG, null);
	}

	public void debug(String context, String logMessage, Throwable exception) {
		logEvent(context, logMessage, LOG_LEVEL_DEBUG, exception);
	}
	
	public void info(String logMessage) {
		logEvent(null, logMessage, LOG_LEVEL_DEBUG, null);
	}
	
	public void info(String logMessage, Throwable exception) {
		logEvent(null, logMessage, LOG_LEVEL_DEBUG, exception);
	}

	public void info(String context, String logMessage) {
		logEvent(context, logMessage, LOG_LEVEL_DEBUG, null);
	}

	public void info(String context, String logMessage, Throwable exception) {
		logEvent(context, logMessage, LOG_LEVEL_DEBUG, exception);
	}
	
	public void log(String logMessage) {
		logEvent(null, logMessage, LOG_LEVEL_INFO, null);
	}
	
	public void log(String logMessage, Throwable exception) {
		logEvent(null, logMessage, LOG_LEVEL_INFO, exception);
	}

	public void log(String context, String logMessage) {
		logEvent(context, logMessage, LOG_LEVEL_INFO, null);
	}

	public void log(String context, String logMessage, Throwable exception) {
		logEvent(context, logMessage, LOG_LEVEL_INFO, exception);
	}
	
	public void warn(String logMessage) {
		logEvent(null, logMessage, LOG_LEVEL_WARN, null);
	}
	
	public void warn(String logMessage, Throwable exception) {
		logEvent(null, logMessage, LOG_LEVEL_WARN, exception);
	}

	public void warn(String context, String logMessage) {
		logEvent(context, logMessage, LOG_LEVEL_WARN, null);
	}

	public void warn(String context, String logMessage, Throwable exception) {
		logEvent(context, logMessage, LOG_LEVEL_WARN, exception);
	}

	public void error(String logMessage) {
		logEvent(null, logMessage, LOG_LEVEL_ERROR, null);
	}
	
	public void error(String logMessage, Throwable exception) {
		logEvent(null, logMessage, LOG_LEVEL_ERROR, exception);
	}

	public void error(String context, String logMessage) {
		logEvent(context, logMessage, LOG_LEVEL_ERROR, null);
	}

	public void error(String context, String logMessage, Throwable exception) {
		logEvent(context, logMessage, LOG_LEVEL_ERROR, exception);
	}

	public boolean isDebugEnabled() {
		return LoggerFactory.isDebugEnabled(clazz);
	}
	
}
