/**
 *  Copyright (c) 2008...2012 Ekkehard Gentz (Rosenheim, Germany) and others.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *  
 *  Contributors:
 *          Ekkehard Gentz     - initial contribution 
 *     
 *   more info: see folder [project]/copyright
 */
package org.appbus.bb.app.logging4bb.logger;

import java.util.Vector;

import org.appbus.bb.app.logging4bb.api.ILogger;

/**
 * The Factory gives you an ILogger for each Class
 * it's recommended in each Class to define
 * private static final ILogger logger = LoggerFactory.getLogger(MyClass.class);
 * then you can log as usual logger.info("bla") etc
 * 
 * @author ekkehard gentz, rosenheim, germany
 *
 */
public class LoggerFactory {
	
	public static int DEFAULT_LOG_LEVEL = ILogger.LOG_LEVEL_INFO;
	
	// each logger can have a collection of classes which should be 
	// logged with debug level
	// independent from default log levelö
	public static Vector DEBUG_CLASSES;
	
	// the application this Logger belongs to
	// if using Lilith all events of one app are displayed in one window
    public static String APP_NAME;
    
    public static long LOG_STORE_ID;
    
	public static boolean init(
			long logStoreId,
			String appName, 
			int defaultLogLevel) {
		LoggerFactory.LOG_STORE_ID = logStoreId;
		LoggerFactory.APP_NAME = appName;
		LoggerFactory.DEFAULT_LOG_LEVEL = defaultLogLevel;
    	System.out.println("Logger initialized for App: "+appName);
        return true;
    }
	
	public static void setDefaultLogLevel(int level){
		if (level >= ILogger.LOG_LEVEL_INFO && level <= ILogger.LOG_LEVEL_ERROR) {
			DEFAULT_LOG_LEVEL = level;
		}
	}

	/**
	 * test if debug is enabled
	 */
	public static boolean isDebugEnabled() {
		return DEFAULT_LOG_LEVEL == ILogger.LOG_LEVEL_DEBUG;
	}
	
	/**
	 * test if debug is enabled for a specific class
	 */
	public static boolean isDebugEnabled(Class clazz) {
		// TODO test if clazz is contained in Vector DEBUG_CLASSES
		return DEFAULT_LOG_LEVEL == ILogger.LOG_LEVEL_DEBUG;
	}

	/**
	 * provides you with a ILogger specific for this class
	 * 
	 * @param clazz
	 * @return
	 */
	public static ILogger getLogger(Class clazz){
		System.out.println("want to get Logger for Class: "+clazz);
		return new Logger(clazz);
	}
		

}
