/**
 *  Copyright (c) 2008...2012 Ekkehard Gentz (Rosenheim, Germany) and others.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *  
 *  Contributors:
 *          Ekkehard Gentz     - initial contribution 
 *     
 *   more info: see folder [project]/copyright
 */
package org.appbus.bb.app.logging4bb.model;

import java.util.Hashtable;

import org.appbus.bb.app.logging4bb.api.ILogger;
import org.appbus.bb.app.logging4bb.logger.Logger;
import org.appbus.bb.app.logging4bb.logger.LoggerFactory;

import net.rim.device.api.i18n.SimpleDateFormat;
import net.rim.device.api.system.Characters;
import net.rim.device.api.system.DeviceInfo;

/**
 * COMMA inside String values is replced by SEMICOLON
 * because COMMA is delimniter for Socket Transmissions
 * 
 * @author ekke
 *
 */
public class LogEventTask {	
	
	private long id;
	
	//private int deviceId;
	
	private long logstoreId;
	private String appName4DB;
	private String appName4Socket;
	private String loggerName;
	private String context;
	private String message;
	private Throwable exception;
	private int level;
	private long eventTime;
	
	private boolean transmittedToServer;
	
	private boolean isInvalid;
	
	//
	private StringBuffer sb;
	private static SimpleDateFormat sdfstamp = 
		new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
	
	private static String emptyString = "#!_!#";
	
	private static String constructAppNameForSocket(String appName){
		StringBuffer sb = new StringBuffer();
		sb.append(appName);
		sb.append(Characters.LEFT_CURLY_BRACKET);
		sb.append(DeviceInfo.getDeviceName());
		sb.append(Characters.COLON);
		sb.append(Integer.toHexString(DeviceInfo.getDeviceId()));
		sb.append(Characters.RIGHT_CURLY_BRACKET);
		return sb.toString();
	}
	
	private static String constructAppNameForDB(String appName){
		StringBuffer sb = new StringBuffer();
		sb.append(appName);
		sb.append(Characters.LEFT_CURLY_BRACKET);
		sb.append(DeviceInfo.getDeviceName());
		sb.append(Characters.RIGHT_CURLY_BRACKET);
		return sb.toString();
	}
	
	private LogEventTask() {
		// dont use it
	}
	
	/**
	 * Constructor used from remote Commands
	 * all attributes are stored into a Hashtable
	 * 
	 * @param remoteLogEvent
	 */
	public LogEventTask(Hashtable remoteLogEvent){
		if (remoteLogEvent == null) {
			setInvalid(true);
			return;
		}
		setInvalid(false);
		this.transmittedToServer = false;
		Object o;
		o = remoteLogEvent.get(ILogger.LOGSTOREID_KEY);
		if (o != null && o instanceof Long) {
			this.logstoreId = ((Long) o).longValue();
		}
		o = remoteLogEvent.get(ILogger.APPNAMEDB_KEY);
		if (o != null && o instanceof String) {
			this.appName4DB = (String)o;
		}
		o = remoteLogEvent.get(ILogger.APPNAMESOCKET_KEY);
		if (o != null && o instanceof String) {
			this.appName4Socket = (String)o;
		}
		o = remoteLogEvent.get(ILogger.LOGGERNAME_KEY);
		if (o != null && o instanceof String) {
			this.loggerName = (String)o;
		}
		o = remoteLogEvent.get(ILogger.CONTEXT_KEY);
		if (o != null && o instanceof String) {
			this.context = (String)o;
		}
		o = remoteLogEvent.get(ILogger.MESSAGE_KEY);
		if (o != null && o instanceof String) {
			this.message = (String)o;
		}
		o = remoteLogEvent.get(ILogger.LEVEL_KEY);
		if (o != null && o instanceof Integer) {
			this.level = ((Integer)o).intValue();
		}
		o = remoteLogEvent.get(ILogger.EVENTTIME_KEY);
		if (o != null && o instanceof Long) {
			this.eventTime = ((Long) o).longValue();
		}
		o = remoteLogEvent.get(ILogger.EXCEPTION_KEY);
		if (o != null && o instanceof Throwable) {
			this.exception = (Throwable) o;
		}
	}
	
	/**
	 * Constructor used from local log events
	 * (events inside this application if eat-your-own-dopgfood is true)
	 * 
	 * @param logstoreId
	 * @param appName
	 * @param loggerName
	 */
	public LogEventTask(long logstoreId, String appName, String loggerName) {
		this.id = Long.MIN_VALUE; // will be autoincremented by SQLite
		this.logstoreId = logstoreId;
		this.appName4DB = constructAppNameForDB(appName);
		this.appName4Socket = constructAppNameForSocket(appName);
		this.loggerName = loggerName;
		this.transmittedToServer = false;
		this.eventTime = Long.MIN_VALUE;
		setInvalid(false);
	}

	public long getId() {
		return id;
	}

	public void setId(long id) {
		this.id = id;
	}

	public int getDeviceId() {
		return DeviceInfo.getDeviceId();
	}
	
	public String getDeviceIdAsHex() {
		return Integer.toHexString(DeviceInfo.getDeviceId());
	}

	public void setMessage(String message) {
		if (message != null && message.length() > 0) {
			this.message = message.replace(Characters.COMMA, Characters.SEMICOLON);
			this.message = this.message.replace(Characters.AMPERSAND, Characters.PLUS_SIGN);
		} else {
			this.message = message;
		}
	}

	public String getMessage() {
		return message;
	}

	public long getLogstoreId() {
		return logstoreId;
	}

	public void setLogstoreId(long logstoreId) {
		this.logstoreId = logstoreId;
	}

	public String getAppName4DB() {
		return appName4DB;
	}

	public void setAppName4DB(String appName) {
		this.appName4DB = appName;
	}
	
	public String getAppName4LogMonitor() {
		return appName4Socket;
	}

	public void setAppName4LogMonitor(String appName) {
		this.appName4Socket = appName;
	}

	public String getLoggerName() {
		return loggerName;
	}

	public void setLoggerName(String loggerName) {
		this.loggerName = loggerName;
	}

	public String getContext() {
		return context;
	}

	public void setContext(String context) {
		if (context != null && context.length() > 0) {
			this.context = context.replace(Characters.COMMA, Characters.SEMICOLON);
		} else {
			this.context = context;
		}
	}

	public Throwable getException() {
		return exception;
	}

	public void setException(Throwable exception) {
		this.exception = exception;
	}

	public int getLevel() {
		return level;
	}

	/**
	 * invalid values are set to DEBUG
	 * @param level
	 */
	public void setLevel(int level) {
		if (level < ILogger.LOG_LEVEL_DEBUG || level > ILogger.LOG_LEVEL_ERROR) {
			this.level = ILogger.LOG_LEVEL_DEBUG;
		} else {
			this.level = level;
		}
	}

	public long getEventTime() {
		return eventTime;
	}

	public void setEventTime(long eventTime) {
		this.eventTime = eventTime;
	}

	public void setTransmittedToServer(boolean transmittedToServer) {
		this.transmittedToServer = transmittedToServer;
	}

	public boolean isTransmittedToServer() {
		return transmittedToServer;
	}

	public void setInvalid(boolean isInvalid) {
		this.isInvalid = isInvalid;
	}

	public boolean isInvalid() {
		return isInvalid;
	}



}
