/**
 *  Copyright (c) 2008...2012 Ekkehard Gentz (Rosenheim, Germany) and others.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *  
 *  Contributors:
 *          Ekkehard Gentz     - initial contribution 
 *     
 *   more info: see folder [project]/copyright
 */
package org.appbus.bb.app.logging4bb.util;

import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;

import org.appbus.bb.app.logging4bb.Logging4BBApp;

import net.rim.device.api.i18n.SimpleDateFormat;
import net.rim.device.api.system.Characters;
import net.rim.device.api.system.LED;

public class Utility {
	
	private static final long HOUR_IN_MILLISECONDS = 3600000L;
	private static final long MINUTES_IN_MILLISECONDS = 60000L;
	private static final long NO_OFFSET = 0L;
	private static String GMT = "GMT";
	
	private static final SimpleDateFormat sdfstamp = 
		new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
	
	public static String getFormattedDate4DB(long timestamp){
		return sdfstamp.formatLocal(timestamp);
	}
	public static String getFormattedDate4DBasTimeZoneGMT(long timestamp, long offsetMilliseconds){
		return sdfstamp.formatLocal(timestamp - offsetMilliseconds);
	}
	
	/**
	 * returns the offset in milliseconds as
	 * GMT +/- hh:mm
	 * 
	 * per ex:
	 * GMT +01:00
	 * GMT -04:30
	 * 
	 * @param offsetMilliseconds
	 * @return
	 */
	public static String getFormattedOffsetToTimeZoneGMT(long offsetMilliseconds){
		StringBuffer sbOffset = new StringBuffer();
		sbOffset.append(GMT);
		sbOffset.append(Characters.SPACE);
		if (offsetMilliseconds < 0) {
			sbOffset.append(Characters.HYPHEN_MINUS);
			offsetMilliseconds *= -1;
		} else {
			sbOffset.append(Characters.PLUS_SIGN);
		}
		long hh = offsetMilliseconds / HOUR_IN_MILLISECONDS;
		if (hh < 10) {
			sbOffset.append(Characters.DIGIT_ZERO);
		}
		sbOffset.append(hh);
		sbOffset.append(Characters.COLON);
		long mm = offsetMilliseconds - (hh * HOUR_IN_MILLISECONDS);
		if (mm > 0) {
			 sbOffset.append(mm / MINUTES_IN_MILLISECONDS);
		} else {
			sbOffset.append(Characters.DIGIT_ZERO);
			sbOffset.append(Characters.DIGIT_ZERO);
		}
		return sbOffset.toString();
	}
	
	/**
	 * calculates the offset in milliseconds
	 * between GMT and local time
	 * for default TimeZone
	 * 
	 * @param timestamp
	 * @return
	 */
	public static long getDaylightOffsetToTimeZoneGMT(long timestamp){
		return getDaylightOffsetToTimeZoneGMT(TimeZone.getDefault(), timestamp);
	}
	
	/**
	 * calculates the offset in milliseconds
	 * between GMT and local time
	 * for TimeZone ID (per ex. 'EUROPE/AMSTERDAM')
	 * if ID was not found, GMT was used == returns 0
	 * 
	 * @param timestamp
	 * @return
	 */
	public static long getDaylightOffsetToTimeZoneGMT(String timeZoneId, long timestamp){
		return getDaylightOffsetToTimeZoneGMT(TimeZone.getTimeZone(timeZoneId), timestamp);
	}
	
	/**
	 * calculates the offset in milliseconds
	 * between GMT and local time
	 * for given TimeZone
	 * 
	 * @param timeZone
	 * @param timestamp
	 * @return
	 */
	public static long getDaylightOffsetToTimeZoneGMT(TimeZone timeZone, long timestamp){
		if (timeZone == null) {
			return NO_OFFSET;
		}
		Calendar calendar = Calendar.getInstance(timeZone);
		calendar.setTime(new Date(timestamp));
		return timeZone.getOffset(
				1, 
				calendar.get(Calendar.YEAR), 
				calendar.get(Calendar.MONTH), 
				calendar.get(Calendar.DAY_OF_MONTH), 
				calendar.get(Calendar.DAY_OF_WEEK), 
				calendar.get(Calendar.MILLISECOND));
	}
	

	// HELPER
	/**
	 * LED mode can be switched on at startup
	 * or later via command
	 */
	public static void useLED(int color, int state) {
		if (!Logging4BBApp.getInstance().isDebugMode()) {
			return;
		}
		try {
			LED.setColorConfiguration(
					500, // onTime, 
					1000, //offTime, 
					color);
		} catch (IllegalArgumentException e) {
			// TODO Auto-generated catch block
		} catch (Throwable e) {
			// TODO: handle exception
		}
		//
		try {
			LED.setState(state);
		} catch (IllegalArgumentException e) {
			// TODO Auto-generated catch block
		} catch (Throwable e) {
			// TODO: handle exception
		}
		//
	}
}
